from __future__ import annotations

import json
from typing import TYPE_CHECKING

from dissect.sql import sqlite3
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

HitmanAlertRecord = TargetRecordDescriptor(
    "application/av/sophos/hitman/log",
    [
        ("datetime", "ts"),
        ("string", "alert"),
        ("string", "description"),
        ("string", "details"),
    ],
)

SophosLogRecord = TargetRecordDescriptor(
    "application/av/sophos/home/log",
    [
        ("datetime", "ts"),
        ("string", "description"),
        ("path", "path"),
    ],
)


class SophosPlugin(Plugin):
    """Sophos antivirus plugin."""

    __namespace__ = "sophos"

    LOG_SOPHOS_HOME = "sysvol/ProgramData/Sophos/Clean/Logs/Clean.log"
    LOG_SOPHOS_HITMAN = "sysvol/ProgramData/HitmanPro.Alert/excalibur.db"
    MARKER_INFECTION = '{"command":"clean-threat'

    LOGS = (LOG_SOPHOS_HOME, LOG_SOPHOS_HITMAN)

    def __init__(self, target: Target):
        super().__init__(target)
        self.codepage = self.target.codepage or "ascii"

    def check_compatible(self) -> None:
        is_compatible = False
        for marker in self.LOGS:
            if self.target.fs.path(marker).exists():
                is_compatible = True
                break
        if not is_compatible:
            raise UnsupportedPluginError("No Sophos/Hitman logs found")

    @export(record=HitmanAlertRecord)
    def hitmanlogs(self) -> Iterator[HitmanAlertRecord]:
        """Return alert log records from Sophos Hitman Pro/Alert.

        Yields HitmanAlertRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp.
            alert (string): Type of Alert.
            description (string): Short description of the alert.
            details (string): Detailed description of the alert.

        Note that because Hitman also catches suspicious behaviour of
        systems, the details field might contain a lot of text, it might
        contain stracktraces etc.
        """
        if self.target.fs.path(self.LOG_SOPHOS_HITMAN).exists():
            try:
                fh = self.target.fs.path(self.LOG_SOPHOS_HITMAN).open("rb")
                db = sqlite3.SQLite3(fh)
                alerts = next(filter(lambda t: t.name == "Alerts", db.tables()))
                for alert in alerts.rows():
                    yield HitmanAlertRecord(
                        ts=wintimestamp(alert.Timestamp),  # already utc
                        alert=alert.AlertType,
                        description=alert.Description,
                        details=alert.Details,
                        _target=self.target,
                    )
            except Exception as e:
                self.target.log.error("Error occurred during reading alerts: %r", e)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

    @export(record=SophosLogRecord)
    def sophoshomelogs(self) -> Iterator[SophosLogRecord]:
        """Return log history records from Sophos Home.

        Yields SophosLogRecord with the following fields:

        .. code-block:: text

            ts (datetime): Timestamp.
            description (string): Short description of the alert.
            path (path): Path to the infected file (if available).

        """
        if self.target.fs.path(self.LOG_SOPHOS_HOME).exists():
            for line in self.target.fs.path(self.LOG_SOPHOS_HOME).open("rt", 0, "utf-16le"):
                if line.find(self.MARKER_INFECTION) > -1:
                    try:
                        ts, json_data = line.split(" ", maxsplit=2)
                        details = json.loads(json_data)

                        path_to_infected_file = None
                        if targets := details.get("targets", None):
                            path_to_infected_file = targets[0].get("file_path", None)

                        # < 3.9.4.1
                        if path := details.get("file_path"):
                            path_to_infected_file = path
                        # > 3.10.3
                        elif targets := details.get("targets", None):
                            path_to_infected_file = targets[0].get("file_path", None)
                            path = (self.target.fs.path(path_to_infected_file),)

                        yield SophosLogRecord(
                            ts=ts,
                            description=details.get("threat_name", details),
                            path=self.target.fs.path(path_to_infected_file),
                            _target=self.target,
                        )
                    except Exception as e:
                        self.target.log.warning("Error: %r on log line: %s", e, line)
                        self.target.log.debug("", exc_info=e)



from __future__ import annotations

from typing import TYPE_CHECKING

from flow.record.fieldtypes import datetime as dt

from dissect.target.plugins.apps.av.sophos import (
    HitmanAlertRecord,
    SophosLogRecord,
    SophosPlugin,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_sophos_hitman_plugin_log(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/sophos/excalibur.db")
    fs_win.map_file("ProgramData/HitmanPro.Alert/excalibur.db", log_file)
    target_win.add_plugin(SophosPlugin)
    logs = list(target_win.sophos.hitmanlogs())
    assert len(logs) == 1
    log = logs[0]
    assert isinstance(log, type(HitmanAlertRecord()))
    assert log.ts == dt("2023-06-14T15:15:53.379862Z")
    assert log.alert == "ILoveYou"
    assert log.description == "Love Letter Virus"
    assert log.details.find("LOVE-LETTER-FOR-YOU.TXT.vbs") > -1


def test_sophos_home_plugin_log_3_9_4_1(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/sophos/Clean-3.9.4.1.log")
    fs_win.map_file("ProgramData/Sophos/Clean/Logs/Clean.log", log_file)
    target_win.add_plugin(SophosPlugin)
    logs = list(target_win.sophos.sophoshomelogs())
    assert len(logs) == 1
    log = logs[0]
    assert log.ts == dt("2023-06-14T10:46:56.235Z")
    assert isinstance(log, type(SophosLogRecord()))
    assert log.description == "EICAR-AV-Test"
    assert str(log.path) == "C:\\eicar_com.zip"


def test_sophos_home_plugin_log_3_10_3(target_win: Target, fs_win: VirtualFilesystem) -> None:
    log_file = absolute_path("_data/plugins/apps/av/sophos/Clean-3.10.3.log")
    fs_win.map_file("ProgramData/Sophos/Clean/Logs/Clean.log", log_file)
    target_win.add_plugin(SophosPlugin)
    logs = list(target_win.sophos.sophoshomelogs())
    assert len(logs) == 1
    log = logs[0]
    assert log.ts == dt("2023-06-14T10:46:56.235000Z")
    assert isinstance(log, type(SophosLogRecord()))
    assert log.description == "EICAR-AV-Test"
    assert str(log.path) == "C:\\eicar_com.zip"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.unix.linux._os import LinuxPlugin

if TYPE_CHECKING:
    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


class DebianPlugin(LinuxPlugin):
    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/etc/network/interfaces") or fs.exists("/etc/debian_version") or fs.exists("/etc/dpkg/"):
                return fs

        return None



from __future__ import annotations

import logging
import stat
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    from impacket.nt_errors import STATUS_NOT_A_DIRECTORY
    from impacket.smb import ATTR_DIRECTORY, SharedFile
    from impacket.smb3structs import (
        FILE_ATTRIBUTE_NORMAL,
        FILE_NON_DIRECTORY_FILE,
        FILE_OPEN,
        FILE_READ_DATA,
        FILE_SHARE_READ,
    )
    from impacket.smbconnection import SessionError, SMBConnection

    HAS_IMPACKET = True
except ImportError:
    HAS_IMPACKET = False

log = logging.getLogger(__name__)


class SmbFilesystem(Filesystem):
    """Filesystem implementation for SMB."""

    __type__ = "smb"

    def __init__(self, conn: SMBConnection, share_name: str, *args, **kwargs):
        super().__init__(None, *args, **kwargs, alt_separator="\\", case_sensitive=False)

        if not HAS_IMPACKET:
            raise ImportError(
                "Required dependency 'impacket' is missing, install with 'pip install dissect.target[smb]'"
            )

        self.conn = conn
        self.share_name = share_name

    @staticmethod
    def detect(fh: BinaryIO) -> bool:
        """No, your file is not an SMB connection."""
        raise TypeError("Detect is not allowed on SmbFilesystem class")

    def get(self, path: str) -> FilesystemEntry:
        """Returns a SmbFilesystemEntry object corresponding to the given path."""
        path = fsutil.normalize(path, self.alt_separator)
        return SmbFilesystemEntry(self, path, self._get_entry(path))

    def _get_entry(self, path: str) -> SharedFile:
        if not path.strip("/"):
            # Getting proper information about the root of the share is cumbersome so just fake it
            return SharedFile(0, 0, 0, 0, 0, ATTR_DIRECTORY, "", "")

        try:
            result = self.conn.listPath(self.share_name, path)
        except SessionError as e:
            if e.error == STATUS_NOT_A_DIRECTORY:
                # STATUS_NOT_A_DIRECTORY
                raise NotADirectoryError(path) from e
            # 0xC000000F is STATUS_NO_SUCH_FILE, but everything else should raise a FileNotFoundError anyway
            raise FileNotFoundError(path) from e

        if len(result) != 1:
            raise FileNotFoundError(path)

        return result[0]


class SmbFilesystemEntry(FilesystemEntry):
    fs: SmbFilesystem
    entry: SharedFile

    def get(self, path: str) -> FilesystemEntry:
        return self.fs.get(fsutil.join(self.path, path, alt_separator=self.fs.alt_separator))

    def _iterdir(self) -> Iterator[SharedFile]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        path = fsutil.join(self.path, "*", alt_separator=self.fs.alt_separator)
        try:
            entry: SharedFile

            for entry in self.fs.conn.listPath(self.fs.share_name, path):
                if entry.get_longname() in (".", ".."):
                    continue

                yield entry
        except SessionError as e:
            log.error("Failed to list directory '%s' share '%s', error: %s", path, self.fs.share_name, e)  # noqa: TRY400

    def iterdir(self) -> Iterator[str]:
        for entry in self._iterdir():
            yield entry.get_longname()

    def scandir(self) -> Iterator[FilesystemEntry]:
        for entry in self._iterdir():
            entry_path = fsutil.join(self.path, entry.get_longname(), alt_separator=self.fs.alt_separator)
            yield SmbFilesystemEntry(self.fs, entry_path, entry)

    def open(self) -> SmbStream:
        log.debug("Attempting to open file: %s", self.path)
        try:
            return SmbStream(self.fs.conn, self.fs.share_name, self.path, self.entry.get_filesize())
        except SessionError as e:
            raise FilesystemError(f"Failed to open file: {self.path}") from e

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return bool(self._resolve(follow_symlinks=follow_symlinks).entry.is_directory())
        except FilesystemError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return not self._resolve(follow_symlinks=follow_symlinks).is_dir()
        except FilesystemError:
            return False

    def is_symlink(self) -> bool:
        return False

    def readlink(self) -> str:
        raise NotASymlinkError

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        mode = stat.S_IFDIR if self.is_dir() else stat.S_IFREG
        st_info = [
            mode | 0o755,
            fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator),
            id(self.fs),
            1,
            0,
            0,
            self.entry.get_filesize(),
            self.entry.get_atime_epoch(),
            self.entry.get_mtime_epoch(),
            self.entry.get_ctime_epoch(),
        ]

        return fsutil.stat_result(st_info)


class SmbStream(AlignedStream):
    """Stream implementation for reading SMB files."""

    def __init__(self, conn: SMBConnection, share_name: str, path: str, size: int):
        self.conn = conn
        self.share_name = share_name
        self.path = path

        self.tree_id = self.conn.connectTree(share_name)
        self.file_id = self.conn.openFile(
            treeId=self.tree_id,
            pathName=path,
            desiredAccess=FILE_READ_DATA,
            shareMode=FILE_SHARE_READ,
            creationOption=FILE_NON_DIRECTORY_FILE,
            creationDisposition=FILE_OPEN,
            fileAttributes=FILE_ATTRIBUTE_NORMAL,
        )
        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        return self.conn.readFile(self.tree_id, self.file_id, offset, length)

    def close(self) -> None:
        try:
            self.conn.closeFile(self.tree_id, self.file_id)
        except Exception as e:
            log.warning("Failed to close file descriptor %d: %s", self.file_id, e)

        try:
            log.debug("Attempting to disconnect tree: %s (id=%d)", self.share_name, self.tree_id)
            self.conn.disconnectTree(self.tree_id)
        except Exception as e:
            log.warning("Failed to disconnect from tree (share=%s, tree_id=%d): %s", self.share_name, self.tree_id, e)



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

import pytest

from dissect.target.loader import open as loader_open
from dissect.target.loaders.acquire import AcquireTarSubLoader, AcquireZipSubLoader
from dissect.target.loaders.tar import TarLoader
from dissect.target.loaders.zip import ZipLoader
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.target import Target
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from pytest_benchmark.fixture import BenchmarkFixture

    from dissect.target.loader import Loader


@pytest.mark.parametrize(
    ("opener"),
    [
        pytest.param(Target.open, id="target-open"),
        pytest.param(lambda x: next(Target.open_all([x])), id="target-open-all"),
    ],
)
def test_target_open(opener: Callable[[str | Path], Target]) -> None:
    """Test that we correctly use ``AcquireTarSubLoader`` when opening a ``Target``."""
    path = absolute_path("_data/loaders/acquire/test-windows-sysvol-absolute.tar")

    target = opener(path)
    assert isinstance(target._loader, TarLoader)
    assert target.path == path


def test_case_sensitive_drive_letter() -> None:
    """Test that we correctly handle case sensitivity of drive letters."""
    path = absolute_path("_data/loaders/acquire/uppercase_driveletter.tar")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, AcquireTarSubLoader)

    # mounts = / and c:
    assert sorted(t.fs.mounts.keys()) == ["c:", "fs"]
    assert "C:" not in t.fs.mounts

    # Initialize our own WindowsPlugin to override the detection
    t._os_plugin = WindowsPlugin.create(t, t.fs.mounts["c:"])
    t.apply()

    # sysvol is now added
    assert sorted(t.fs.mounts.keys()) == ["c:", "fs", "sysvol"]

    # WindowsPlugin sets the case sensitivity to False
    assert t.fs.get("C:/test.file").open().read() == b"hello_world"
    assert t.fs.get("c:/test.file").open().read() == b"hello_world"


@pytest.mark.parametrize(
    ("archive", "expected_drive_letter"),
    [
        ("_data/loaders/acquire/test-windows-sysvol-absolute.tar", "c:"),  # C: due to backwards compatibility
        ("_data/loaders/acquire/test-windows-sysvol-relative.tar", "c:"),  # C: due to backwards compatibility
        ("_data/loaders/acquire/test-windows-fs-c-relative.tar", "c:"),
        ("_data/loaders/acquire/test-windows-fs-c-absolute.tar", "c:"),
        ("_data/loaders/acquire/test-windows-fs-x.tar", "x:"),
    ],
)
def test_windows_sysvol_formats(archive: str, expected_drive_letter: str) -> None:
    """Test that we correctly handle different sysvol formats."""
    path = absolute_path(archive)

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, AcquireTarSubLoader)

    assert WindowsPlugin.detect(t)
    # NOTE: for the sysvol archives, this also tests the backwards compatibility
    assert sorted(t.fs.mounts.keys()) == [expected_drive_letter]
    assert t.fs.get(f"{expected_drive_letter}/Windows/System32/foo.txt")


def test_windows_sysvol_formats_zip() -> None:
    """Test that we correctly handle sysvol formats in ZIP archives."""
    path = absolute_path("_data/loaders/acquire/test-windows-fs-c.zip")

    loader = loader_open(path)
    assert isinstance(loader, ZipLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, AcquireZipSubLoader)

    assert WindowsPlugin.detect(t)
    # NOTE: for the sysvol archives, this also tests the backwards compatibility
    assert sorted(t.fs.mounts.keys()) == ["c:"]
    assert t.fs.get("c:/Windows/System32/foo.txt")


def test_anonymous_filesystems() -> None:
    """Test that we correctly handle anonymous filesystems in Acquire archives."""
    path = absolute_path("_data/loaders/acquire/test-anon-filesystems.tar")

    loader = loader_open(path)
    assert isinstance(loader, TarLoader)

    t = Target()
    loader.map(t)
    assert isinstance(loader.subloader, AcquireTarSubLoader)

    assert t.fs.get("$fs$/fs0/foo").open().read() == b"hello world\n"
    assert t.fs.get("$fs$/fs1/bar").open().read() == b"hello world\n"


@pytest.mark.parametrize(
    ("archive", "loader"),
    [
        ("_data/loaders/acquire/test-windows-fs-c-relative.tar", TarLoader),
        ("_data/loaders/acquire/test-windows-fs-c.zip", ZipLoader),
    ],
)
@pytest.mark.benchmark
def test_benchmark(benchmark: BenchmarkFixture, archive: str, loader: type[Loader]) -> None:
    """Benchmark the loading of Acquire archives."""
    file = absolute_path(archive)

    benchmark(lambda: loader(file).map(Target()))



from __future__ import annotations

import hashlib
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.os.windows.rdpcache import (
    BMP_DATA_OFFSET,
    BMP_MAGIC,
    BitmapTile,
    RdpCachePlugin,
    assemble_tiles_into_collage,
    c_bmp,
    extract_bin,
    extract_bmc,
    parse_color_data,
    tile_to_bitmap,
    wrap_square_colors_in_border,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target import Target
    from dissect.target.filesystem import VirtualFilesystem

BMC_PATH = absolute_path("_data/plugins/os/windows/rdpcache/bcache24.bmc")
BIN_PATH = absolute_path("_data/plugins/os/windows/rdpcache/Cache0000.bin")


@pytest.fixture
def target_win_rdp_cache(target_win_users: Target, fs_win: VirtualFilesystem) -> Target:
    fs_win.map_file("Users\\John\\AppData\\Local\\Microsoft\\Terminal Server Client\\Cache\\Cache0000.bin", BIN_PATH)
    fs_win.map_file("Users\\John\\AppData\\Local\\Microsoft\\Terminal Server Client\\Cache\\bcache24.bmc", BMC_PATH)

    target_win_users.add_plugin(RdpCachePlugin)

    return target_win_users


def test_wrap_in_border() -> None:
    """Test if ``rdpcache.wrap_square_colors_in_border`` behaves as expected."""

    grey_pixel = b"\x80\x80\x80\xff"
    blue_pixel = b"\xff\x00\x00\xff"
    grey_square = grey_pixel * 4  # A 2x2 square of grey pixels

    bordered_square = wrap_square_colors_in_border(grey_square, 2, blue_pixel, 2)

    # The square should now be 6x6 (2 lines added above, 2 on the left, 2 on the right, 2 on the bottom)
    assert len(bordered_square) == 6 * 6 * 4  # 4 bytes per pixel
    assert bordered_square.startswith(blue_pixel * 6)
    assert bordered_square.endswith(blue_pixel * 6)

    # Grab the third 'row' of the square (6 pixels horizontally, and 4 bytes per pixel)
    stripe = bordered_square[(6 * 3 * 4) : (6 * 4 * 4)]
    assert stripe == (blue_pixel * 2) + (grey_pixel * 2) + (blue_pixel * 2)


def test_bmc_no_remnants() -> None:
    """Test if ``rdpcache.extract_bmc`` behaves as expected."""

    with BMC_PATH.open("rb") as fh:
        tiles = list(extract_bmc(fh))

        # The test sample contains 'remnant tiles', but these contain no color data (its all null bytes)
        # We expect the parser to filter those out
        assert len(tiles) == 40
        assert not any(tile.is_remnant for tile in tiles)

        total_tile_data = b"".join(tile.colors for tile in tiles)
        assert hashlib.md5(total_tile_data).hexdigest() == "3fb7c485c7ee83e0d5d748d2e1c9d206"


def test_bin() -> None:
    """Test if ``rdpcache.extract_bin`` behaves as expected."""

    with BIN_PATH.open("rb") as fh:
        tiles = list(extract_bin(fh))

        assert len(tiles) == 254
        assert not any(tile.is_remnant for tile in tiles)
        total_tile_data = b"".join(tile.colors for tile in tiles)
        assert hashlib.md5(total_tile_data).hexdigest() == "7e7a88aa54efd92b3ab8e4f7b29afe3f"


def test_collage() -> None:
    """Test ``rdpcache.assemble_tiles_into_collage`` behaves as expected."""

    with BIN_PATH.open("rb") as fh:
        tiles = list(extract_bin(fh))
        collage_no_borders = assemble_tiles_into_collage(tiles)
        assert collage_no_borders.width == 64 * 64  # Rows should be max. 64 tiles
        assert collage_no_borders.height == 64 * 4  # 4
        assert hashlib.md5(collage_no_borders.colors).hexdigest() == "eabbe8a19e0d88ba15483eea0785dd39"

        collage_with_borders = assemble_tiles_into_collage(tiles, border_around_tile=2)
        assert hashlib.md5(collage_with_borders.colors).hexdigest() == "dbcd1181fd362570ae7b232aaeb136fa"
        assert collage_with_borders.width == 68 * 64  # Rows should be max. 64 tiles
        assert collage_with_borders.height == 68 * 4  # 47 rows


def test_bmp_export() -> None:
    """Test if we can convert a tile to a bitmap correctly."""

    grey_pixel = b"\x80\x80\x80\xff"
    square = grey_pixel * 16
    tile = BitmapTile(4, 4, square)

    bmp = tile_to_bitmap(tile)
    assert bmp == bytes.fromhex(
        "424dba000000000000007a0000006c00000004000000040000000100200003000000400000000000"
        "00000000000000000000000000000000ff0000ff0000ff000000000000ff206e6957000000000000"
        "00000000000000000000000000000000000000000000000000000000000000000000000000000000"
        "0000808080ff808080ff808080ff808080ff808080ff808080ff808080ff808080ff808080ff8080"
        "80ff808080ff808080ff808080ff808080ff808080ff808080ff"
    )

    bmp_fh = BytesIO(bmp)

    file_header = c_bmp.BITMAPFILEHEADER(bmp_fh)
    assert file_header.bfType == BMP_MAGIC
    assert file_header.bfOffBits == BMP_DATA_OFFSET
    bitmap_header = c_bmp.BITMAPV4HEADER(bmp_fh)
    assert bitmap_header.bV4Width == 4
    assert bitmap_header.bV4Height == 4
    assert bitmap_header.bV4SizeImage == 16 * 4  # 16 pixels of 4 bytes

    assert bmp_fh.read() == square  # The rest of the bitmap should be the color data


def test_parse_color_data() -> None:
    """Test ``rdpcache.parse_color_data`` behavior."""

    blue_color_half_transparency = b"\xff\x00\x00\x80"
    blue_square = blue_color_half_transparency * 4

    # The bmc parsing boosts transparency, replacing every fourth byte with 255
    assert parse_color_data(blue_square) == b"\xff\x00\x00\xff" * 4

    blue_pixel = b"\xff\x00\x00\xff"
    red_pixel = b"\x00\x00\xff\xff"

    blue_row = blue_pixel * 64
    red_row = red_pixel * 64

    # For BIN files, the tiles are saved 'top down' but we want to save them 'bottom up' later, as that is the BMP
    # default
    assert parse_color_data(blue_row + red_row, True, 64) == red_row + blue_row

    # While tiles are usually of width 64, they're not all like that. We want to be sure that color data isn't picked
    # up in chunks of 64 pixels only.
    blue_row = blue_pixel * 32
    red_row = red_pixel * 32
    assert parse_color_data(blue_row + red_row, True, 32) == red_row + blue_row


def test_rdp_cache_plugin(target_win_rdp_cache: Target, tmp_path: Path) -> None:
    """Test if the ``rdpcache.recover`` and ``rdpcache.paths`` behave as expected."""

    # Create a directory to extract the cache to
    cache_dst = tmp_path.joinpath("rdp_cache")
    cache_dst.mkdir()

    target_win_rdp_cache.rdpcache.recover(
        output_dir=cache_dst, no_individual_tiles=False, as_collage=True, as_grid=True, remnants=True
    )

    prefix = f"{target_win_rdp_cache.name}_John_"
    assert cache_dst.joinpath(f"{prefix}Cache0000.bin_grid.bmp").exists()
    assert cache_dst.joinpath(f"{prefix}Cache0000.bin_collage.bmp").exists()
    assert cache_dst.joinpath(f"{prefix}bcache24.bmc_grid.bmp").exists()
    assert cache_dst.joinpath(f"{prefix}bcache24.bmc_collage.bmp").exists()

    assert len(list(cache_dst.glob(f"{prefix}Cache0000.bin_*"))) == 256  # 254 tiles
    assert len(list(cache_dst.glob(f"{prefix}bcache24.bmc_*"))) == 42  # 39 tiles

    records = list(target_win_rdp_cache.rdpcache.paths())
    assert len(records) == 2
    assert records[0].path == "C:\\Users\\John\\AppData\\Local\\Microsoft\\Terminal Server Client\\Cache\\Cache0000.bin"
    assert records[1].path == "C:\\Users\\John\\AppData\\Local\\Microsoft\\Terminal Server Client\\Cache\\bcache24.bmc"



from __future__ import annotations

from dissect.target.plugins.apps.edr.acquire import AcquirePlugin
from dissect.target.target import Target
from tests._utils import absolute_path


def test_acquire_handles_plugin() -> None:
    file_hashes_target = Target().open(absolute_path("_data/plugins/apps/edr/acquire/handles/test-acquire-handles.tar"))
    file_hashes_target.add_plugin(AcquirePlugin)

    results = list(file_hashes_target.acquire.handles())
    first_result = results[0]

    assert first_result.name == r"\Windows\Fonts"
    assert first_result.handle_type == "EtwRegistration"
    assert first_result.unique_process_id == 1
    assert first_result.object == "0xfffftest"
    assert results[-1].unique_process_id == 124
    assert len(results) == 124


def test_acquire_hash_plugin() -> None:
    file_hashes_target = Target().open(absolute_path("_data/plugins/apps/edr/acquire/hash/test-acquire-hash.tar"))
    file_hashes_target.add_plugin(AcquirePlugin)

    results = list(file_hashes_target.acquire.hashes())

    assert results[0].path == "/sysvol/Windows/bfsvc.exe"
    assert len(results) == 998



from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING
from xml.etree.ElementTree import ParseError

from defusedxml import ElementTree
from flow.record import GroupedRecord

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target import Target
    from dissect.target.helpers.fsutil import TargetPath


RecentlyUsedRecord = TargetRecordDescriptor(
    "unix/linux/recently_used",
    [
        ("datetime", "ts"),
        ("string", "user"),
        ("string", "source"),
        ("string", "href"),
        ("datetime", "added"),
        ("datetime", "modified"),
        ("datetime", "visited"),
        ("string", "mimetype"),
        ("string", "groups"),
        ("boolean", "private"),
    ],
)

RecentlyUsedIconRecord = TargetRecordDescriptor(
    "unix/linux/recently_used/icon",
    [
        ("string", "type"),
        ("string", "href"),
        ("string", "name"),
    ],
)

RecentlyUsedApplicationRecord = TargetRecordDescriptor(
    "unix/linux/recently_used/application",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "exec"),
        ("varint", "count"),
    ],
)
ns = {
    "bookmark": "http://www.freedesktop.org/standards/desktop-bookmarks",
    "mime": "http://www.freedesktop.org/standards/shared-mime-info",
}


def parse_ts(target: Target, ts: str) -> datetime | None:
    """Parse timestamp format from xbel file

    Returns None if unable to parse the timestamp"""
    if ts is None:
        return None

    try:
        # datetime.fromisoformat() doesn't support the trailing Z in python <= 3.10
        return datetime.strptime(ts, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
    except ValueError as e:
        # Use None if we're unable to parse the timestamp
        target.log.warning("Could not parse timestamp %s, using None instead", ts)
        target.log.debug("", exc_info=e)
        return None


def parse_recently_used_xbel(
    target: Target, username: str, xbel_file: TargetPath
) -> Iterator[RecentlyUsedRecord | GroupedRecord] | None:
    with xbel_file.open() as fh:
        try:
            et = ElementTree.fromstring(fh.read(), forbid_dtd=True)
        except ParseError as e:
            target.log.warning("Could not parse %s, skipping", xbel_file)
            target.log.debug("", exc_info=e)
            return

        for bookmark in et.iter("bookmark"):
            # The spec says there should always be exactly one.
            # Ignore if there are fewer or more.
            mimetypes = bookmark.findall("./info/metadata/mime:mime-type", ns)
            if mimetypes and len(mimetypes) == 1:  # noqa SIM108
                mimetype = mimetypes[0].get("type")
            else:
                mimetype = None

            # This is just a list of names, GroupedRecords seem overkill
            groups = bookmark.findall("./info/metadata/bookmark:groups/bookmark:group", ns)
            group_list = ", ".join(group.text for group in groups)

            # There should be at most one "private" tag, but accept multiple
            private_entries = bookmark.findall("./info/metadata/bookmark:private", ns)
            private = len(private_entries) > 0

            cur = RecentlyUsedRecord(
                ts=parse_ts(target, bookmark.get("visited")),
                user=username,
                source=xbel_file,
                href=bookmark.get("href"),
                added=parse_ts(target, bookmark.get("added")),
                modified=parse_ts(target, bookmark.get("modified")),
                visited=parse_ts(target, bookmark.get("visited")),
                mimetype=mimetype,
                groups=group_list,
                private=private,
                _target=target,
            )
            yield cur

            # Icon is optional, spec says at most one.
            for icon in bookmark.findall("./info/metadata/bookmark:icon", ns):
                iconrecord = RecentlyUsedIconRecord(
                    type=icon.get("type"),
                    href=icon.get("href"),
                    name=icon.get("name"),
                )
                yield GroupedRecord("unix/linux/recently_used/icon", [cur, iconrecord])

            # Spec says there should be at least one application
            for app in bookmark.findall("./info/metadata/bookmark:applications/bookmark:application", ns):
                apprecord = RecentlyUsedApplicationRecord(
                    ts=parse_ts(target, app.get("modified")),
                    name=app.get("name"),
                    exec=app.get("exec"),
                    count=app.get("count"),
                )
                yield GroupedRecord("unix/linux/recently_used/application", [cur, apprecord])


class RecentlyUsedPlugin(Plugin):
    """Parse recently-used.xbel files on Gnome-based Linux Desktops.

    Based on the spec on https://www.freedesktop.org/wiki/Specifications/desktop-bookmark-spec/
    """

    FILEPATH = ".local/share/recently-used.xbel"

    def __init__(self, target: Target):
        super().__init__(target)
        self.users_files: list[tuple[str, TargetPath]] = []
        for user_details in self.target.user_details.all_with_home():
            xbel_file = user_details.home_path.joinpath(self.FILEPATH)
            if not xbel_file.exists():
                continue
            self.users_files.append((user_details.user, xbel_file))

    def check_compatible(self) -> None:
        if not len(self.users_files):
            raise UnsupportedPluginError("No recently-used.xbel files found")

    @export(record=RecentlyUsedRecord)
    def recently_used(self) -> Iterator[RecentlyUsedRecord | GroupedRecord]:
        """Parse recently-used.xbel files on Linux Desktops."""

        for user, xbel_file in self.users_files:
            yield from parse_recently_used_xbel(self.target, user.name, xbel_file)



from __future__ import annotations

import gzip
import io
import logging
import os
import pathlib
import stat
from collections import defaultdict
from typing import TYPE_CHECKING, Any, BinaryIO, Callable, Final

from dissect.target.exceptions import (
    FileNotFoundError,
    FilesystemError,
    IsADirectoryError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.target.helpers import fsutil, hashutil
from dissect.target.helpers.lazy import import_lazy

TarFilesystem = import_lazy("dissect.target.filesystems.tar").TarFilesystem

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

FILESYSTEMS: list[type[Filesystem]] = []
MODULE_PATH = "dissect.target.filesystems"

log = logging.getLogger(__name__)


class Filesystem:
    """Base class for filesystems."""

    # Due to lazy importing we generally can't use isinstance(), so we add a short identifying string to each class
    # This has the added benefit of having a readily available "pretty name" for each implementation
    __type__: str = None
    """A short string identifying the type of filesystem."""
    __multi_volume__: bool = False
    """Whether this filesystem supports multiple volumes (disks)."""

    def __init__(
        self,
        volume: BinaryIO | list[BinaryIO] | None = None,
        alt_separator: str = "",
        case_sensitive: bool = True,
    ) -> None:
        """The base initializer for the class.

        Args:
            volume: A volume or other file-like object associated with the filesystem.
            case_sensitive: Defines if the paths in the filesystem are case sensitive or not.
            alt_separator: The alternative separator used to distingish between directories in a path.

        Raises:
            NotImplementedError: When the internal ``__type__`` of the class is not defined.
        """
        self.volume = volume
        self.case_sensitive = case_sensitive
        self.alt_separator = alt_separator

        if self.__type__ is None:
            raise NotImplementedError(f"{self.__class__.__name__} must define __type__")

    def __repr__(self) -> str:
        return f"<Filesystem type={self.__type__}>"

    def path(self, *args) -> fsutil.TargetPath:
        """Instantiate a new path-like object on this filesystem."""
        return fsutil.TargetPath(self, *args)

    @classmethod
    def detect(cls, fh: BinaryIO) -> bool:
        """Detect whether the ``fh`` file-handle is supported by this ``Filesystem`` implementation.

        The position of ``fh`` will be restored before returning.

        Args:
            fh: A file-like object, usually a disk or partition.

        Returns:
            ``True`` if ``fh`` is supported, ``False`` otherwise.
        """
        offset = fh.tell()
        try:
            fh.seek(0)
            return cls._detect(fh)
        except NotImplementedError:
            raise
        except Exception as e:
            log.warning("Failed to detect %s filesystem", cls.__type__)
            log.debug("", exc_info=e)
        finally:
            fh.seek(offset)

        return False

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        """Detect whether the ``fh`` file-handle is supported by this ``Filesystem`` implementation.

        This method should be implemented by subclasses. The position of ``fh`` is guaranteed to be ``0``.

        Args:
            fh: A file-like object, usually a disk or partition.

        Returns:
            ``True`` if ``fh`` is supported, ``False`` otherwise.
        """
        raise NotImplementedError

    @classmethod
    def detect_id(cls, fh: BinaryIO) -> bytes | None:
        """Return a filesystem set identifier.

        Only used in filesystems that support multiple volumes (disks) to find all volumes
        belonging to a single filesystem.

        Args:
            fh: A file-like object, usually a disk or partition.
        """
        if not cls.__multi_volume__:
            return None

        offset = fh.tell()
        try:
            fh.seek(0)
            return cls._detect_id(fh)
        except NotImplementedError:
            raise
        except Exception as e:
            log.warning("Failed to detect ID on %s filesystem", cls.__type__)
            log.debug("", exc_info=e)
        finally:
            fh.seek(offset)

        return None

    @staticmethod
    def _detect_id(fh: BinaryIO) -> bytes | None:
        """Return a filesystem set identifier.

        This method should be implemented by subclasses of filesystems that support multiple volumes (disks).
        The position of ``fh`` is guaranteed to be ``0``.

        Args:
            fh: A file-like object, usually a disk or partition.

        Returns:
            An identifier that can be used to combine the given ``fh`` with others beloning to the same set.
        """
        raise NotImplementedError

    def iter_subfs(self) -> Iterator[Filesystem]:
        """Yield possible sub-filesystems."""
        yield from ()

    def get(self, path: str) -> FilesystemEntry:
        """Retrieve a :class:`FilesystemEntry` from the filesystem.

        Args:
            path: The path which we want to retrieve.

        Returns:
            A :class:`FilesystemEntry` for the path.
        """
        raise NotImplementedError

    def open(self, path: str) -> BinaryIO:
        """Open a filesystem entry.

        Args:
            path: The location on the filesystem to open.

        Returns:
            A file-like object. Resolves symlinks when possible.
        """
        return self.get(path).open()

    def iterdir(self, path: str) -> Iterator[str]:
        """Iterate over the contents of a directory, return them as strings.

        Args:
            path: The location on the filesystem to iterate over.

        Returns:
            An iterator of directory entries as path strings.
        """
        return self.get(path).iterdir()

    def scandir(self, path: str) -> Iterator[FilesystemEntry]:
        """Iterate over the contents of a directory, return them as FilesystemEntry's.

        Args:
            path: The directory to scan.

        Returns:
            An iterator of directory entries as FilesystemEntry's.
        """
        return self.get(path).scandir()

    def listdir(self, path: str) -> list[str]:
        """List the contents of a directory as strings.

        Args:
            path: The directory to get the listing from.

        Returns:
            A list of path strings.
        """
        return list(self.iterdir(path))

    def listdir_ext(self, path: str) -> list[FilesystemEntry]:
        """List the contents of a directory as FilesystemEntry's.

        Args:
            path: The directory to get the listing from.

        Returns:
            A list of FilesystemEntry's.
        """
        return list(self.scandir(path))

    def walk(
        self,
        path: str,
        topdown: bool = True,
        onerror: Callable[[Exception], None] | None = None,
        followlinks: bool = False,
    ) -> Iterator[tuple[str, list[str], list[str]]]:
        """Recursively walk a directory pointed to by ``path``, returning the string representation of both files
        and directories.

        Args:
            path: The path to walk on the filesystem.
            topdown: ``True`` puts the ``path`` at the top, ``False`` puts the ``path`` at the bottom.
            onerror: A method to execute when an error occurs.
            followlinks: ``True`` if we want to follow any symbolic link.

        Returns:
            An iterator of directory entries as path strings.
        """
        return self.get(path).walk(topdown, onerror, followlinks)

    def walk_ext(
        self,
        path: str,
        topdown: bool = True,
        onerror: Callable[[Exception], None] | None = None,
        followlinks: bool = False,
    ) -> Iterator[tuple[list[FilesystemEntry], list[FilesystemEntry], list[FilesystemEntry]]]:
        """Recursively walk a directory pointed to by ``path``, returning :class:`FilesystemEntry` of files
        and directories.

        Args:
            path: The path to walk on the filesystem.
            topdown: ``True`` puts the ``path`` at the top, ``False`` puts the ``path`` at the bottom.
            onerror: A method to execute when an error occurs.
            followlinks: ``True`` if we want to follow any symbolic link.

        Returns:
            An iterator of directory entries as FilesystemEntry's.
        """
        return self.get(path).walk_ext(topdown, onerror, followlinks)

    def recurse(self, path: str) -> Iterator[FilesystemEntry]:
        """Recursively walk a directory and yield contents as :class:`FilesystemEntry`.

        Does not follow symbolic links.

        Args:
            path: The path to recursively walk on the target filesystem.

        Returns:
            An iterator of :class:`FilesystemEntry`.
        """
        return self.get(path).recurse()

    def glob(self, pattern: str) -> Iterator[str]:
        """Iterate over the directory part of ``pattern``, returning entries matching ``pattern`` as strings.

        Args:
            pattern: The pattern to match.

        Returns:
            An iterator of path strings that match the pattern.
        """
        for entry in self.glob_ext(pattern):
            yield entry.path

    def glob_ext(self, pattern: str) -> Iterator[FilesystemEntry]:
        """Iterate over the directory part of ``pattern``, returning entries matching ``pattern`` as FilesysmteEntry's.


        Args:
            pattern: The pattern to match.

        Returns:
            An iterator of FilesystemEntry's that match the pattern.
        """
        path, pattern = fsutil.glob_split(pattern, alt_separator=self.alt_separator)
        try:
            entry = self.get(path)
        except FileNotFoundError:
            return
        else:
            yield from fsutil.glob_ext(entry, pattern)

    def exists(self, path: str) -> bool:
        """Determines whether ``path`` exists on a filesystem.

        If the ``path`` is a symbolic link, it will attempt to resolve it to find the FilesystemEntry it points to.

        Args:
            path: a path on the filesystem.

        Returns:
            ``True`` if the given path exists, ``False`` otherwise.
        """
        try:
            entry = self.get(path)
            if entry.is_symlink():
                entry.readlink_ext()
        except Exception:
            return False
        else:
            return True

    def lexists(self, path: str) -> bool:
        """Determines if a ``path`` exists on the filesystem without resolving links.

        Args:
            path: A path on the filesystem.

        Returns:
            ``True`` if the given path is a file, ``False`` otherwise.
        """
        try:
            self.get(path)
        except Exception:
            return False
        else:
            return True

    def is_file(self, path: str, follow_symlinks: bool = True) -> bool:
        """Determine if ``path`` is a file on the filesystem.

        Args:
            path: The path on the filesystem.
            follow_symlinks: Whether to resolve the path if it is a symbolic link.

        Returns:
            ``True`` if the given path is a file or a symbolic link to a file, return ``False`` otherwise.  If
            ``follow_symlinks`` is ``False``, return ``True`` only if the given path is a file (without
            following symlinks).
        """
        try:
            return self.get(path).is_file(follow_symlinks=follow_symlinks)
        except FileNotFoundError:
            return False

    def is_dir(self, path: str, follow_symlinks: bool = True) -> bool:
        """Determine whether the given ``path`` is a directory on the filesystem.

        Args:
            path: The path on the filesystem.
            follow_symlinks: Whether to resolve the path if it is a symbolic link.

        Returns:
            ``True`` if the given path is a directory or a symbolic link to a directory, return ``False``
            otherwise.
            If ``follow_symlinks`` is ``False``, return ``True`` only if the given path is a directory
            (without following symlinks).
        """
        try:
            return self.get(path).is_dir(follow_symlinks=follow_symlinks)
        except FileNotFoundError:
            return False

    def is_symlink(self, path: str) -> bool:
        """Determine wether the given ``path`` is a symlink on the filesystem.

        Args:
            path: The path on the filesystem.

        Returns:
            ``True`` if the given path is a symbolic link, ``False`` otherwise.
        """
        try:
            return self.get(path).is_symlink()
        except FileNotFoundError:
            return False

    def readlink(self, path: str) -> str:
        """Read the link where the given ``path`` points to, return the resulting path as string.

        If it is a symlink and returns the string that corresponds to that path.
        This means it follows the path a link points to, it tries to do it recursively.

        Args:
            path: The symbolic link to read.

        Returns:
            The path the link points to.
        """
        return self.get(path).readlink()

    def readlink_ext(self, path: str) -> FilesystemEntry:
        """Read the link where the given ``path`` points to, return the resulting path as FilesystemEntry.

        If it is a symlink and returns the entry that corresponds to that path.
        This means it follows the path a link points to, it tries to do it recursively.

        Args:
            path: The symbolic link to read.

        Returns:
            The ``FilesystemEntry`` where the symbolic link points to.
        """
        return self.get(path).readlink_ext()

    def stat(self, path: str, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Determine the stat information of a ``path`` on the filesystem.

        If ``path`` is a symlink and ``follow_symlinks`` is ``True``, it gets resolved, attempting to stat the
        path where it points to.

        Args:
            path: The filesystem path we want the stat information from.
            follow_symlinks: Whether to resolve the path if it is a symbolic link.

        Returns:
            The stat information of the given path.
        """
        return self.get(path).stat(follow_symlinks=follow_symlinks)

    def lstat(self, path: str) -> fsutil.stat_result:
        """Determine the stat information of a ``path`` on the filesystem, **without** resolving symlinks.

        When it detects a symlink, it will stat the information of the symlink, not the path it points to.

        Args:
            path: The filesystem path we want the stat information from.

        Returns:
            The stat information of the given path.
        """
        return self.get(path).lstat()

    def md5(self, path: str) -> str:
        """Calculate the MD5 digest of the contents of the file ``path`` points to.

        Args:
            path: The filesystem path to get the digest from.

        Returns:
            The MD5 digest of the contents of ``path``.
        """
        return self.get(path).md5()

    def sha1(self, path: str) -> str:
        """Calculate the SHA1 digest of the contents of the file ``path`` points to.

        Args:
            path: The filesystem path to get the digest from.

        Returns:
            The SHA1 digest of the contents of ``path``.
        """
        return self.get(path).sha1()

    def sha256(self, path: str) -> str:
        """Calculate the SHA256 digest of the contents of the file ``path`` points to.

        Args:
            path: The filesystem path to get the digest from.

        Returns:
            The SHA256 digest of the contents of ``path``.
        """
        return self.get(path).sha256()

    def hash(self, path: str, algos: list[str] | list[Callable] | None = None) -> tuple[str]:
        """Calculate the digest of the contents of ``path``, using the ``algos`` algorithms.

        Args:
            path: The filesystem path to get the digest from.
            algos: The types of hashes to calculate. If ``None`` it will use the common set of algorithms defined in
                        :py:func:`dissect.target.helpers.hashutil.common` as ``[MD5, SHA1, SHA256]``.

        Returns:
            The digests of the contents of ``path``.
        """
        return self.get(path).hash(algos)


class FilesystemEntry:
    """Base class for filesystem entries."""

    def __init__(self, fs: Filesystem, path: str, entry: Any):
        """Initialize the base filesystem entry class.

        Args:
            fs: The filesystem to get data from.
            path: The path of the entry mapped on ``fs``.
            entry: The raw entry backing this filesystem entry.
        """
        self.fs = fs
        self.path = path
        self.name = fsutil.basename(path, alt_separator=self.fs.alt_separator)
        self.entry = entry

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.path!r}>"

    def __str__(self) -> str:
        return str(self.path)

    def _resolve(self, follow_symlinks: bool = True) -> FilesystemEntry:
        """Helper method to resolve symbolic links.

        If ``follow_symlinks`` is ``False``, this function is effectively a no-op.

        Args:
            follow_symlinks: Whether to resolve the entry if it is a symbolic link.

        Returns:
            The resolved symbolic link if ``follow_symlinks`` is ``True`` and the :class:`FilesystemEntry` is a
            symbolic link or else the :class:`FilesystemEntry` itself.
        """
        if follow_symlinks and self.is_symlink():
            return self.readlink_ext()
        return self

    def get(self, path: str) -> FilesystemEntry:
        """Retrieve a :class:`FilesystemEntry` relative to this entry.

        Args:
            path: The path relative to this filesystem entry.

        Returns:
            A relative :class:`FilesystemEntry`.
        """
        raise NotImplementedError

    def open(self) -> BinaryIO:
        """Open this filesystem entry.

        Returns:
            A file-like object. Resolves symlinks when possible
        """
        raise NotImplementedError

    def iterdir(self) -> Iterator[str]:
        """Iterate over the contents of a directory, return them as strings.

        Returns:
            An iterator of directory entries as path strings.
        """
        raise NotImplementedError

    def scandir(self) -> Iterator[FilesystemEntry]:
        """Iterate over the contents of a directory, yields :class:`FilesystemEntry`.

        Returns:
            An iterator of :class:`FilesystemEntry`.
        """
        raise NotImplementedError

    def listdir(self) -> list[str]:
        """List the contents of a directory as strings.

        Returns:
            A list of path strings.
        """
        return list(self.iterdir())

    def listdir_ext(self) -> list[FilesystemEntry]:
        """List the contents of a directory as a list of :class:`FilesystemEntry`.

        Returns:
            A list of :class:`FilesystemEntry`.
        """
        return list(self.scandir())

    def walk(
        self,
        topdown: bool = True,
        onerror: Callable[[Exception], None] | None = None,
        followlinks: bool = False,
    ) -> Iterator[tuple[str, list[str], list[str]]]:
        """Recursively walk a directory and yield its contents as strings split in a tuple
        of lists of files, directories and symlinks.

        These contents include::
          -files
          -directories
          -symboliclinks

        Args:
            topdown: ``True`` puts this entry at the top of the list, ``False`` puts this entry at the bottom.
            onerror: A method to execute when an error occurs.
            followlinks: ``True`` if we want to follow any symbolic link.

        Returns:
            An iterator of directory entries as path strings.
        """
        yield from fsutil.walk(self, topdown, onerror, followlinks)

    def walk_ext(
        self,
        topdown: bool = True,
        onerror: Callable[[Exception], None] | None = None,
        followlinks: bool = False,
    ) -> Iterator[tuple[list[FilesystemEntry], list[FilesystemEntry], list[FilesystemEntry]]]:
        """Recursively walk a directory and yield its contents as :class:`FilesystemEntry` split in a tuple of
        lists of files, directories and symlinks.

        Args:
            topdown: ``True`` puts this entry at the top of the list, ``False`` puts this entry at the bottom.
            onerror: A method to execute when an error occurs.
            followlinks: ``True`` if we want to follow any symbolic link

        Returns:
            An iterator of tuples :class:`FilesystemEntry`.
        """
        yield from fsutil.walk_ext(self, topdown, onerror, followlinks)

    def recurse(self) -> Iterator[FilesystemEntry]:
        """Recursively walk a directory and yield its contents as :class:`FilesystemEntry`.

        Does not follow symbolic links.

        Returns:
            An iterator of :class:`FilesystemEntry`.
        """
        yield from fsutil.recurse(self)

    def glob(self, pattern: str) -> Iterator[str]:
        """Iterate over this directory part of ``patern``, returning entries matching ``pattern`` as strings.

        Args:
            pattern: The pattern to match.

        Returns:
            An iterator of path strings that match the pattern.
        """
        for entry in self.glob_ext(pattern):
            yield entry.path

    def glob_ext(self, pattern: str) -> Iterator[FilesystemEntry]:
        """Iterate over the directory part of ``pattern``, returning entries matching
        ``pattern`` as :class:`FilesysmteEntry`.

        Args:
            pattern: The pattern to glob for.

        Returns:
            An iterator of :class:`FilesystemEntry` that match the pattern.
        """
        yield from fsutil.glob_ext(self, pattern)

    def exists(self, path: str) -> bool:
        """Determines whether a ``path``, relative to this entry, exists.

        If the `path` is a symbolic link, it will attempt to resolve it to find
        the :class:`FilesystemEntry` it points to.

        Args:
            path: The path relative to this entry.

        Returns:
            ``True`` if the path exists, ``False`` otherwise.
        """
        try:
            entry = self.get(path)
            if entry.is_symlink():
                entry.readlink_ext()
        except Exception:
            return False
        else:
            return True

    def lexists(self, path: str) -> bool:
        """Determine wether a ``path`` relative to this enty, exists without resolving links.

        Args:
            path: The path relative to this entry.

        Returns:
            ``True`` if the path exists, ``False`` otherwise.
        """
        try:
            self.get(path)
        except Exception:
            return False
        else:
            return True

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Determine if this entry is a file.

        Args:
            follow_symlinks: Whether to resolve the entry if it is a symbolic link.

        Returns:
            ``True`` if the entry is a file or a symbolic link to a file, return ``False`` otherwise.
            If ``follow_symlinks`` is ``False``, return ``True`` only if the entry is a file (without
            following symlinks).
        """
        raise NotImplementedError

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Determine if this entry is a directory.

        Args:
            follow_symlinks: Whether to resolve the entry if it is a symbolic link.

        Returns:
            ``True`` if the entry is a directory or a symbolic link to a directory, return ``False``
            otherwise.
            If ``follow_symlinks`` is ``False``, return ``True`` only if the entry is a directory (without
            following symlinks).

        """
        raise NotImplementedError

    def is_symlink(self) -> bool:
        """Determine whether this entry is a symlink.

        Returns:
            ``True`` if the entry is a symbolic link, ``False`` otherwise.
        """
        raise NotImplementedError

    def readlink(self) -> str:
        """Read the link where this entry points to, return the resulting path as string.

        If it is a symlink and returns the entry that corresponds to that path.
        This means it follows the path a link points to, it tries to do it recursively.

        Returns:
            The path the link points to."""
        raise NotImplementedError

    def readlink_ext(self) -> FilesystemEntry:
        """Read the link where this entry points to, return the resulting path as :class:`FilesystemEntry`.

        If it is a symlink and returns the string that corresponds to that path.
        This means it follows the path a link points to, it tries to do it recursively.

        Returns:
            The filesystem entry the link points to.
        """
        log.debug("%r::readlink_ext()", self)
        # Default behavior, resolve link own filesystem.
        return fsutil.resolve_link(self.fs, self.readlink(), self.path, alt_separator=self.fs.alt_separator)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        """Determine the stat information of this entry.

        If the entry is a symlink and ``follow_symlinks`` is ``True``, it gets resolved, attempting to stat
        the path where it points to.

        Args:
            follow_symlinks: Whether to resolve the symbolic link if this entry is a symbolic link.

        Returns:
            The stat information of this entry.
        """
        raise NotImplementedError

    def lstat(self) -> fsutil.stat_result:
        """Determine the stat information of this entry, **without** resolving the symlinks.

        When it detects a symlink, it will stat the information of the symlink, not the path it points to.

        Returns:
            The stat information of this entry.
        """
        raise NotImplementedError

    def attr(self) -> Any:
        """The attributes related to this entry, resolving any symlinks.

        If the entry is a symbolic link, it will attempt to resolve it first.
        Resulting in the attr information of the entry it points to.

        Returns:
            The attributes of this entry.
        """
        raise NotImplementedError

    def lattr(self) -> Any:
        """The attributes related to this current entry, **without** resolving links.

        Returns:
            The attributes of this entry.
        """
        raise NotImplementedError

    def md5(self) -> str:
        """Calculates the MD5 digest of this entry.

        Returns:
            The MD5 digest of this entry.
        """
        return hashutil.md5(self.open())

    def sha1(self) -> str:
        """Calculates the SHA1 digest of this entry.

        Returns:
            The SHA1 digest of this entry.
        """
        return hashutil.sha1(self.open())

    def sha256(self) -> str:
        """Calculates the SHA256 digest of this entry.

        Returns:
            The SHA256 digest of this entry.
        """
        return hashutil.sha256(self.open())

    def hash(self, algos: list[str] | list[Callable] | None = None) -> tuple[str]:
        """Calculate the digest of this entry, using the ``algos`` algorithms.

        Args:
            algos: The types of hashes to calculate. If ``None`` it will use the common set of algorithms defined in
                   :py:func:`dissect.target.helpers.hashutil.common` as ``[MD5, SHA1, SHA256]``.

        Returns:
            The various digests of this entry.
        """
        if algos:
            return hashutil.custom(self.open(), algos)
        return hashutil.common(self.open())


class VirtualDirectory(FilesystemEntry):
    """Virtual directory implementation. Backed by a dict."""

    def __init__(self, fs: VirtualFilesystem, path: str):
        super().__init__(fs, path, None)
        self.up = None
        self.top = None
        self.entries = {}

    def __getitem__(self, item: str) -> FilesystemEntry:
        if not self.fs.case_sensitive:
            item = item.lower()
        return self.entries[item]

    def __contains__(self, item: str) -> bool:
        if not self.fs.case_sensitive:
            item = item.lower()
        return item in self.entries

    def open(self) -> BinaryIO:
        raise IsADirectoryError(f"{self.path} is a directory")

    def attr(self) -> Any:
        raise TypeError(f"attr is not allowed on VirtualDirectory: {self.path}")

    def lattr(self) -> Any:
        raise TypeError(f"lattr is not allowed on VirtualDirectory: {self.path}")

    def add(self, name: str, entry: FilesystemEntry) -> None:
        """Add an entry to this :class:`VirtualDirectory`."""
        if not self.fs.case_sensitive:
            name = name.lower()

        self.entries[name] = entry

    def get(self, path: str) -> FilesystemEntry:
        return self.fs.get(path, relentry=self)

    def iterdir(self) -> Iterator[str]:
        yielded = set()
        for entry in self.entries:
            yield entry
            yielded.add(entry)

        # self.top used to be a reference to a filesystem. This is now a reference to
        # any filesystem entry, usually the root of a filesystem.
        if self.top:
            for entry in self.top.iterdir():
                if entry in yielded or (not self.fs.case_sensitive and entry.lower() in yielded):
                    continue
                yield entry

    def scandir(self) -> Iterator[FilesystemEntry]:
        yielded = set()
        for entry in self.entries.values():
            yield entry
            yielded.add(entry.name)

        # self.top used to be a reference to a filesystem. This is now a reference to
        # any filesystem entry, usually the root of a filesystem.
        if self.top:
            for entry in self.top.scandir():
                if entry.name in yielded or (not self.fs.case_sensitive and entry.name.lower() in yielded):
                    continue
                yield entry

    def _stat(self) -> fsutil.stat_result:
        path_addr = fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator)
        return fsutil.stat_result([stat.S_IFDIR, path_addr, id(self.fs), 1, 0, 0, 0, 0, 0, 0])

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        if self.top:
            return self.top.stat(follow_symlinks=follow_symlinks)
        return self._stat()

    def lstat(self) -> fsutil.stat_result:
        if self.top:
            return self.top.lstat()
        return self._stat()

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        return True

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return False

    def is_symlink(self) -> bool:
        return False

    def readlink(self) -> str:
        raise NotASymlinkError(self.path)

    def readlink_ext(self) -> FilesystemEntry:
        raise NotASymlinkError(self.path)


class VirtualFileHandle(io.RawIOBase):
    def __init__(self, fh: BinaryIO):
        self.fh = fh
        self.seek(0)

    def readinto(self, b: bytearray) -> int:
        return self.fh.readinto(b)

    def seek(self, offset: int, whence: int = io.SEEK_SET) -> int:
        return self.fh.seek(offset, whence)

    def readable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True


class VirtualFile(FilesystemEntry):
    """Virtual file backed by a file-like object."""

    def attr(self) -> Any:
        raise TypeError(f"attr is not allowed on {self.__class__.__name__}: {self.path}")

    def lattr(self) -> Any:
        raise TypeError(f"lattr is not allowed on {self.__class__.__name__}: {self.path}")

    def get(self, path: str) -> FilesystemEntry:
        path = fsutil.normalize(path, alt_separator=self.fs.alt_separator).strip("/")
        if not path:
            return self
        raise NotADirectoryError(f"'{self.path}' is not a directory")

    def iterdir(self) -> Iterator[str]:
        raise NotADirectoryError(f"'{self.path}' is not a directory")

    def scandir(self) -> Iterator[FilesystemEntry]:
        raise NotADirectoryError(f"'{self.path}' is not a directory")

    def open(self) -> BinaryIO:
        return VirtualFileHandle(self.entry)

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        size = getattr(self.entry, "size", 0)
        file_addr = fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator)
        return fsutil.stat_result([stat.S_IFREG, file_addr, id(self.fs), 1, 0, 0, size, 0, 0, 0])

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        return True

    def is_symlink(self) -> bool:
        return False

    def readlink(self) -> str:
        raise NotASymlinkError(self.path)

    def readlink_ext(self) -> FilesystemEntry:
        raise NotASymlinkError(self.path)


class MappedFile(VirtualFile):
    """Virtual file backed by a file on the host machine."""

    entry: str

    def open(self) -> BinaryIO:
        return pathlib.Path(self.entry).open("rb")

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        # Python 3.9 does not support follow_symlinks in stat()
        if follow_symlinks:
            return fsutil.stat_result.copy(pathlib.Path(self.entry).stat())

        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        return fsutil.stat_result.copy(pathlib.Path(self.entry).lstat())

    def attr(self) -> dict[str, bytes]:
        return fsutil.fs_attrs(self.entry, follow_symlinks=True)

    def lattr(self) -> dict[str, bytes]:
        return fsutil.fs_attrs(self.entry, follow_symlinks=False)


class MappedCompressedFile(MappedFile):
    """Virtual file backed by a compressed file on the host machine."""

    entry: str
    _compressors: Final[dict[str, Any]] = {"gzip": gzip}

    def __init__(self, fs: Filesystem, path: str, entry: str, algo: str = "gzip"):
        super().__init__(fs, path, entry)
        self._compressor = self._compressors.get(algo)
        if self._compressor is None:
            raise ValueError(f"Unsupported compression algorithm {algo}")

    def open(self) -> BinaryIO:
        return self._compressor.open(self.entry, "rb")


class VirtualSymlink(FilesystemEntry):
    """Virtual symlink implementation."""

    def __init__(self, fs: Filesystem, path: str, target: str):
        super().__init__(fs, path, None)
        self.target = target

    def attr(self) -> Any:
        return self.readlink_ext().attr()

    def lattr(self) -> Any:
        raise TypeError(f"lattr is not allowed on VirtualSymlink: {self.path}")

    def get(self, path: str) -> FilesystemEntry:
        return self.fs.get(path, self)

    def iterdir(self) -> Iterator[str]:
        yield from self.readlink_ext().iterdir()

    def scandir(self) -> Iterator[FilesystemEntry]:
        yield from self.readlink_ext().scandir()

    def open(self) -> BinaryIO:
        return self.readlink_ext().open()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks).lstat()

    def lstat(self) -> fsutil.stat_result:
        link_addr = fsutil.generate_addr(self.path, alt_separator=self.fs.alt_separator)
        return fsutil.stat_result([stat.S_IFLNK, link_addr, id(self.fs), 1, 0, 0, len(self.target), 0, 0, 0])

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        if not follow_symlinks:
            return False

        try:
            return self.readlink_ext().is_dir()
        except FileNotFoundError:
            return False

    def is_file(self, follow_symlinks: bool = True) -> bool:
        if not follow_symlinks:
            return False

        try:
            return self.readlink_ext().is_file()
        except FileNotFoundError:
            return False

    def is_symlink(self) -> bool:
        return True

    def readlink(self) -> str:
        return self.target


class VirtualFilesystem(Filesystem):
    __type__ = "virtual"

    def __init__(self, **kwargs):
        super().__init__(None, **kwargs)
        self.root = VirtualDirectory(self, "/")

    @staticmethod
    def detect(fh: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on VirtualFilesystem class")

    def get(self, path: str, relentry: FilesystemEntry | None = None) -> FilesystemEntry:
        entry = relentry or self.root
        path = fsutil.normalize(path, alt_separator=self.alt_separator).strip("/")
        full_path = fsutil.join(entry.path, path, alt_separator=self.alt_separator)

        if not path:
            return entry

        parts = path.split("/")

        for i, part in enumerate(parts):
            # If the entry of the previous part (or the starting relentry /
            # root entry) is a symlink, resolve it first so things like entry.up
            # work if it is a symlink to a directory.
            # Note that this will never resolve the final part of the path if
            # that happens to be a symlink, so things like fs.is_symlink() will
            # work.
            if entry.is_symlink():
                entry = entry.readlink_ext()

            if not entry.is_dir():
                # An entry for the current part can only be retrieved if the
                # entry of the previous part (or the starting relentry / root
                # entry) is a directory.
                raise NotADirectoryError(full_path)
            if part == ".":
                continue
            if part == "..":
                entry = entry.up
                if not entry:
                    entry = self.root
            else:
                if part in entry:
                    entry = entry[part]
                elif entry.top:
                    try:
                        return entry.top.get(fsutil.join(*parts[i:], alt_separator=self.alt_separator))
                    except FilesystemError as e:
                        raise FileNotFoundError(full_path) from e
                else:
                    raise FileNotFoundError(full_path)

        return entry

    def makedirs(self, path: str) -> VirtualDirectory:
        """Create virtual directories into the VFS from the given path."""
        path = fsutil.normalize(path, alt_separator=self.alt_separator).strip("/")
        directory = self.root

        if not path:
            return directory

        parts = path.split("/")
        for i, part in enumerate(parts):
            if part not in directory:
                vdir = VirtualDirectory(self, fsutil.join(*parts[: i + 1], alt_separator=self.alt_separator))
                vdir.up = directory

                directory.add(part, vdir)

            directory = directory[part]

        return directory

    def map_fs(self, vfspath: str, fs: Filesystem, base: str = "/") -> None:
        """Mount a dissect filesystem to a directory in the VFS."""
        directory = self.makedirs(vfspath)
        directory.top = fs.get(base)

    mount = map_fs

    def map_dir(self, vfspath: str, realpath: pathlib.Path | str) -> None:
        """Recursively map a directory from the host machine into the VFS."""
        if not isinstance(realpath, pathlib.Path):
            realpath = pathlib.Path(realpath)

        vfspath = fsutil.normalize(vfspath, alt_separator=self.alt_separator).strip("/")
        base = realpath.resolve()

        # Compatibility walk until we are base 3.12
        def _walk_compat(path: pathlib.Path) -> Iterator[tuple[pathlib.Path, list[str], list[str]]]:
            """Compatibility function to walk a directory."""
            for root, dirs, files in os.walk(path):
                yield pathlib.Path(root), dirs, files

        for root, dirs, files in base.walk() if hasattr(base, "walk") else _walk_compat(base):
            relroot = str(root.relative_to(base))
            if relroot == ".":
                relroot = ""

            relroot = fsutil.normalize(relroot, alt_separator=os.path.sep)
            vfsroot = fsutil.join(vfspath, relroot, alt_separator=self.alt_separator)
            directory = self.makedirs(vfsroot)

            for dir_ in dirs:
                vfs_dir = fsutil.join(vfsroot, dir_, alt_separator=self.alt_separator)
                self.makedirs(vfs_dir)

            for file_ in files:
                vfs_file_path = fsutil.join(vfsroot, file_, alt_separator=self.alt_separator)
                real_file_path = root.joinpath(file_)
                directory.add(file_, MappedFile(self, vfs_file_path, str(real_file_path)))

    def map_file(self, vfspath: str, realpath: str, compression: str | None = None) -> None:
        """Map a file from the host machine into the VFS."""
        vfspath = fsutil.normalize(vfspath, alt_separator=self.alt_separator)
        if vfspath[-1] == "/":
            raise AttributeError(f"Can't map a file onto a directory: {vfspath}")
        file_path = vfspath.lstrip("/")

        if compression is not None:
            mapped_file = MappedCompressedFile(self, file_path, realpath, algo=compression)
        else:
            mapped_file = MappedFile(self, file_path, realpath)
        self.map_file_entry(vfspath, mapped_file)

    def map_file_fh(self, vfspath: str, fh: BinaryIO) -> None:
        """Map a file handle into the VFS."""
        vfspath = fsutil.normalize(vfspath, alt_separator=self.alt_separator)
        if vfspath[-1] == "/":
            raise AttributeError(f"Can't map a file onto a directory: {vfspath}")
        file_path = vfspath.lstrip("/")
        self.map_file_entry(vfspath, VirtualFile(self, file_path, fh))

    def map_file_entry(self, vfspath: str, entry: FilesystemEntry) -> None:
        """Map a :class:`FilesystemEntry` into the VFS.

        Any missing subdirectories up to, but not including, the last part of
        ``vfspath`` will be created.
        """
        vfspath = fsutil.normalize(vfspath, alt_separator=self.alt_separator).strip("/")
        if not vfspath:
            self.root.top = entry
        else:
            if "/" in vfspath:
                sub_dirs = fsutil.dirname(vfspath, alt_separator=self.alt_separator)
                directory = self.makedirs(sub_dirs)
            else:
                directory = self.root

            entry_name = fsutil.basename(vfspath, alt_separator=self.alt_separator)
            directory.add(entry_name, entry)

    def map_dir_from_tar(self, vfspath: str, tar_file: str | pathlib.Path, map_single_file: bool = False) -> None:
        """Map files in a tar onto the VFS.

        Args:
            vfspath: Destination path in the virtual filesystem.
            tar_file: Source path of the tar file to map.
            map_single_file: Only mount a single file found inside the tar at the specified path.
        """

        if not isinstance(tar_file, pathlib.Path):
            try:
                tar_file = pathlib.Path(tar_file)
            except TypeError:
                raise ValueError("tar_file should be a string or Path instance")

        vfspath = fsutil.normalize(vfspath, alt_separator=self.alt_separator)
        tfs = TarFilesystem(tar_file.open("rb"))

        if map_single_file:
            # We map the first file we find in the tar to the provided vfspath.
            for file in [f[0] for _, _, f in tfs.walk_ext("/") if f]:
                file.name = fsutil.basename(vfspath)
                self.map_file_entry(vfspath, file)
                return
        else:
            self.mount(vfspath, tfs)

    def map_file_from_tar(self, vfspath: str, tar_file: str | pathlib.Path) -> None:
        """Map a single file in a tar archive to the given path in the VFS.

        The provided tar archive should contain *one* file.

        Args:
            vfspath: Destination path in the virtual filesystem.
            tar_file: Source path of the tar file to map.
        """
        return self.map_dir_from_tar(vfspath.lstrip("/"), tar_file, map_single_file=True)

    def link(self, src: str, dst: str) -> None:
        """Hard link a :class:`FilesystemEntry` to another location.

        Args:
            src: The path to the target of the link.
            dst: The path to the link.
        """
        self.map_file_entry(dst, self.get(src))

    def symlink(self, src: str, dst: str) -> None:
        """Create a symlink to another location.

        Args:
            src: The path to the target of the symlink.
            dst: The path to the symlink.
        """
        src = fsutil.normalize(src, alt_separator=self.alt_separator).rstrip("/")
        dst = fsutil.normalize(dst, alt_separator=self.alt_separator).strip("/")
        self.map_file_entry(dst, VirtualSymlink(self, dst, src))


class LayerFilesystem(Filesystem):
    __type__ = "layer"

    def __init__(self, **kwargs):
        self.layers: list[Filesystem] = []
        self.mounts = {}
        self._alt_separator = "/"
        self._case_sensitive = True
        self._root_entry = LayerFilesystemEntry(self, "/", [])
        self.root = self.append_layer()
        super().__init__(None, **kwargs)

    def __getattr__(self, attr: str) -> Any:
        """Provide "magic" access to filesystem specific attributes from any of the layers.

        For example, on a :class:`LayerFilesystem` ``fs``, you can do ``fs.ntfs`` to access the
        internal NTFS object if it has an NTFS layer.
        """
        for fs in self.layers:
            try:
                return getattr(fs, attr)
            except AttributeError:  # noqa: PERF203
                continue
        else:
            return object.__getattribute__(self, attr)

    @staticmethod
    def detect(fh: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on LayerFilesystem class")

    def mount(self, path: str, fs: Filesystem, ignore_existing: bool = True) -> None:
        """Mount a filesystem at a given path.

        If there's an overlap with an existing mount, creates a new layer.

        Args:
            path: The path to mount the filesystem at.
            fs: The filesystem to mount.
            ignore_existing: Whether to ignore existing mounts and create a new layer. Defaults to ``True``.
        """
        root = self.root
        for mount in self.mounts:
            if ignore_existing and path == mount:
                continue

            if path.startswith(mount):
                root = self.append_layer()
                break

        root.map_fs(path, fs)
        self.mounts[path] = fs

    def link(self, dst: str, src: str) -> None:
        """Hard link a :class:`FilesystemEntry` to another location."""
        self.root.map_file_entry(dst, self.get(src))

    def symlink(self, dst: str, src: str) -> None:
        """Create a symlink to another location."""
        self.root.symlink(dst, src)

    def append_layer(self, **kwargs) -> VirtualFilesystem:
        """Append a new layer."""
        layer = VirtualFilesystem(case_sensitive=self.case_sensitive, alt_separator=self.alt_separator, **kwargs)
        self.append_fs_layer(layer)
        return layer

    add_layer = append_layer

    def prepend_layer(self, **kwargs) -> VirtualFilesystem:
        """Prepend a new layer."""
        layer = VirtualFilesystem(case_sensitive=self.case_sensitive, alt_separator=self.alt_separator, **kwargs)
        self.prepend_fs_layer(layer)
        return layer

    def append_fs_layer(self, fs: Filesystem) -> None:
        """Append a filesystem as a layer.

        Args:
            fs: The filesystem to append.
        """
        # Counterintuitively, we prepend the filesystem to the list of layers
        # We could reverse the list of layers upon iteration, but that is a hot path
        self.layers.insert(0, fs)
        self._root_entry.entries.insert(0, fs.get("/"))

    def prepend_fs_layer(self, fs: Filesystem) -> None:
        """Prepend a filesystem as a layer.

        Args:
            fs: The filesystem to prepend.
        """
        # Counterintuitively, we append the filesystem to the list of layers
        # We could reverse the list of layers upon iteration, but that is a hot path
        self.layers.append(fs)
        self._root_entry.entries.append(fs.get("/"))

    def remove_fs_layer(self, fs: Filesystem) -> None:
        """Remove a filesystem layer.

        Args:
            fs: The filesystem to remove.
        """
        self.remove_layer(self.layers.index(fs))

    def remove_layer(self, idx: int) -> None:
        """Remove a layer by index.

        Args:
            idx: The index of the layer to remove.
        """
        del self.layers[idx]
        del self._root_entry.entries[idx]

    @property
    def case_sensitive(self) -> bool:
        """Whether the filesystem is case sensitive."""
        return self._case_sensitive

    @property
    def alt_separator(self) -> str:
        """The alternative separator of the filesystem."""
        return self._alt_separator

    @case_sensitive.setter
    def case_sensitive(self, value: bool) -> None:
        """Set the case sensitivity of the filesystem (and all layers)."""
        self._case_sensitive = value
        self.root.case_sensitive = value
        for layer in self.layers:
            layer.case_sensitive = value

    @alt_separator.setter
    def alt_separator(self, value: str) -> None:
        """Set the alternative separator of the filesystem (and all layers)."""
        self._alt_separator = value
        self.root.alt_separator = value
        for layer in self.layers:
            layer.alt_separator = value

    def get(self, path: str, relentry: LayerFilesystemEntry | None = None) -> LayerFilesystemEntry:
        """Get a :class:`FilesystemEntry` from the filesystem."""
        entry = relentry or self._root_entry
        path = fsutil.normalize(path, alt_separator=self.alt_separator).strip("/")
        full_path = fsutil.join(entry.path, path, alt_separator=self.alt_separator)

        if not path:
            return entry

        exc = []
        entries = []

        for sub_entry in entry.entries:
            try:
                entries.append(self._get_from_entry(path, sub_entry))
            except FilesystemError as e:  # noqa: PERF203
                exc.append(e)

        if not entries:
            if all(isinstance(ex, NotADirectoryError) for ex in exc):
                raise NotADirectoryError(full_path)
            if all(isinstance(ex, NotASymlinkError) for ex in exc):
                raise NotASymlinkError(full_path)
            raise FileNotFoundError(full_path)

        return LayerFilesystemEntry(self, full_path, entries)

    def _get_from_entry(self, path: str, entry: FilesystemEntry) -> FilesystemEntry:
        """Get a :class:`FilesystemEntry` relative to a specific entry."""
        parts = path.split("/")

        for i, part in enumerate(parts):
            if entry.is_symlink():
                # Resolve using the RootFilesystem instead of the entry's Filesystem
                entry = fsutil.resolve_link(
                    self,
                    entry.readlink(),
                    "/".join(parts[:i]),
                    alt_separator=entry.fs.alt_separator,
                )
            entry = entry.get(part)

        return entry


class EntryList(list):
    """Wrapper list for filesystem entries.

    Exposes a ``__getattr__`` on a list of items. Useful to access internal objects from filesystem implementations.
    For example, access the underlying NTFS object from a list of virtual and NTFS entries.
    """

    def __init__(self, value: FilesystemEntry | list[FilesystemEntry]):
        if not isinstance(value, list):
            value = [value]
        super().__init__(value)

    def __getattr__(self, attr: str) -> Any:
        for entry in self:
            try:
                return getattr(entry, attr)
            except AttributeError:  # noqa: PERF203
                continue
        else:
            return object.__getattribute__(self, attr)


class LayerFilesystemEntry(FilesystemEntry):
    def __init__(self, fs: Filesystem, path: str, entry: FilesystemEntry):
        super().__init__(fs, path, EntryList(entry))
        self.entries: EntryList = self.entry
        self._link = None

    def _exec(self, func: str, *args, **kwargs) -> Any:
        """Helper method to execute a method over all contained entries."""
        exc = []
        for entry in self.entries:
            try:
                return getattr(entry, func)(*args, **kwargs)
            except (AttributeError, NotImplementedError) as e:  # noqa: PERF203
                exc.append(str(e))

        exceptions = ",".join(exc) if exc else "No entries"

        raise FilesystemError(f"Can't resolve {func} for {self}: {exceptions}")

    def get(self, path: str) -> FilesystemEntry:
        return self.fs.get(path, self._resolve())

    def open(self) -> BinaryIO:
        return self._resolve()._exec("open")

    def iterdir(self) -> Iterator[str]:
        yielded = {".", ".."}
        selfentry = self._resolve()
        for fsentry in selfentry.entries:
            for entry_name in fsentry.iterdir():
                name = entry_name if selfentry.fs.case_sensitive else entry_name.lower()
                if name in yielded:
                    continue

                yield entry_name
                yielded.add(name)

    def scandir(self) -> Iterator[LayerFilesystemEntry]:
        # Every entry is actually a list of entries from the different
        # overlaying FSes, of which each may implement a different function
        # like .stat() or .open()
        items = defaultdict(list)
        selfentry = self._resolve()
        for fsentry in selfentry.entries:
            for entry in fsentry.scandir():
                name = entry.name if selfentry.fs.case_sensitive else entry.name.lower()
                if name in (".", ".."):
                    continue

                items[name].append(entry)

        for entries in items.values():
            # The filename for the first entry is taken. Note that in case of
            # non case-sensitive FSes, the different entries from the
            # overlaying FSes may have different casing of the name.
            entry_name = entries[0].name
            path = fsutil.join(selfentry.path, entry_name, alt_separator=selfentry.fs.alt_separator)
            yield LayerFilesystemEntry(selfentry.fs, path, entries)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks)._exec("is_file", follow_symlinks=follow_symlinks)
        except FileNotFoundError:
            return False

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        try:
            return self._resolve(follow_symlinks=follow_symlinks)._exec("is_dir", follow_symlinks=follow_symlinks)
        except FileNotFoundError:
            return False

    def is_symlink(self) -> bool:
        return self._exec("is_symlink")

    def readlink(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError(f"Not a link: {self}")
        return self._exec("readlink")

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self._resolve(follow_symlinks=follow_symlinks)._exec("stat", follow_symlinks=follow_symlinks)

    def lstat(self) -> fsutil.stat_result:
        return self._exec("lstat")

    def attr(self) -> Any:
        return self._resolve()._exec("attr")

    def lattr(self) -> Any:
        return self._exec("lattr")


class RootFilesystem(LayerFilesystem):
    __type__ = "root"

    def __init__(self, target: Target):
        self.target = target
        super().__init__()

    @staticmethod
    def detect(fh: BinaryIO) -> bool:
        raise TypeError("Detect is not allowed on RootFilesystem class")

    def get(self, path: str, relentry: LayerFilesystemEntry | None = None) -> RootFilesystemEntry:
        self.target.log.debug("%r::get(%r)", self, path)
        entry = super().get(path, relentry)
        entry.__class__ = RootFilesystemEntry
        return entry


class RootFilesystemEntry(LayerFilesystemEntry):
    fs: RootFilesystem

    def get(self, path: str) -> RootFilesystemEntry:
        self.fs.target.log.debug("%r::get(%r)", self, path)
        entry = super().get(path)
        entry.__class__ = RootFilesystemEntry
        return entry

    def open(self) -> BinaryIO:
        self.fs.target.log.debug("%r::open()", self)
        return super().open()

    def iterdir(self) -> Iterator[str]:
        self.fs.target.log.debug("%r::iterdir()", self)
        yield from super().iterdir()

    def scandir(self) -> Iterator[RootFilesystemEntry]:
        self.fs.target.log.debug("%r::scandir()", self)
        for entry in super().scandir():
            entry.__class__ = RootFilesystemEntry
            yield entry

    def is_file(self, follow_symlinks: bool = True) -> bool:
        self.fs.target.log.debug("%r::is_file()", self)
        return super().is_file(follow_symlinks=follow_symlinks)

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        self.fs.target.log.debug("%r::is_dir()", self)
        return super().is_dir(follow_symlinks=follow_symlinks)

    def is_symlink(self) -> bool:
        self.fs.target.log.debug("%r::is_symlink()", self)
        return super().is_symlink()

    def readlink(self) -> str:
        self.fs.target.log.debug("%r::readlink()", self)
        return super().readlink()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        self.fs.target.log.debug("%r::stat()", self)
        return super().stat(follow_symlinks=follow_symlinks)

    def lstat(self) -> fsutil.stat_result:
        self.fs.target.log.debug("%r::lstat()", self)
        return super().lstat()

    def attr(self) -> Any:
        self.fs.target.log.debug("%r::attr()", self)
        return super().attr()

    def lattr(self) -> Any:
        self.fs.target.log.debug("%r::lattr()", self)
        return super().lattr()


def register(module: str, class_name: str, internal: bool = True) -> None:
    """Register a filesystem implementation to use when opening a filesystem.

    This function registers a filesystem using ``module`` relative to the ``MODULE_PATH``.
    It lazily imports the module, and retrieves the specific class from it.

    Args:
        module: The module where to find the filesystem.
        class_name: The class to load.
        internal: Whether it is an internal module or not.
    """

    if internal:
        module = f"{MODULE_PATH}.{module}"

    FILESYSTEMS.append(getattr(import_lazy(module), class_name))


def is_multi_volume_filesystem(fh: BinaryIO) -> bool:
    for filesystem in FILESYSTEMS:
        try:
            if filesystem.__multi_volume__ and filesystem.detect(fh):
                return True
        except ImportError as e:  # noqa: PERF203
            log.info("Failed to import %s", filesystem)
            log.debug("", exc_info=e)

    return False


def open(fh: BinaryIO, *args, **kwargs) -> Filesystem:
    offset = fh.tell()
    fh.seek(0)

    try:
        for filesystem in FILESYSTEMS:
            try:
                if filesystem.detect(fh):
                    return filesystem(fh, *args, **kwargs)
            except ImportError as e:  # noqa: PERF203
                log.info("Failed to import %s", filesystem)
                log.debug("", exc_info=e)
            except Exception as e:
                raise FilesystemError(f"Failed to open filesystem for {fh}") from e
    finally:
        fh.seek(offset)

    raise FilesystemError(f"Failed to open filesystem for {fh}")


def open_multi_volume(fhs: list[BinaryIO], *args, **kwargs) -> Iterator[Filesystem]:
    for filesystem in FILESYSTEMS:
        try:
            if not filesystem.__multi_volume__:
                continue

            volumes = defaultdict(list)
            for fh in fhs:
                if not filesystem.detect(fh):
                    continue

                identifier = filesystem.detect_id(fh)
                volumes[identifier].append(fh)

            for vols in volumes.values():
                yield filesystem(vols, *args, **kwargs)

        except ImportError as e:
            log.info("Failed to import %s", filesystem)
            log.debug("", exc_info=e)


register("ntfs", "NtfsFilesystem")
register("extfs", "ExtFilesystem")
register("xfs", "XfsFilesystem")
register("fat", "FatFilesystem")
register("ffs", "FfsFilesystem")
register("vmfs", "VmfsFilesystem")
register("btrfs", "BtrfsFilesystem")
register("exfat", "ExfatFilesystem")
register("squashfs", "SquashFSFilesystem")
register("jffs", "JffsFilesystem")
register("qnxfs", "QnxFilesystem")
register("zip", "ZipFilesystem")
register("tar", "TarFilesystem")
register("vmtar", "VmtarFilesystem")
register("cpio", "CpioFilesystem")
register("vbk", "VbkFilesystem")
register("ad1", "AD1Filesystem")



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.loaders.vmx import VmxLoader

if TYPE_CHECKING:
    from pathlib import Path


class VmwarevmLoader(VmxLoader):
    """Load ``*.vmwarevm`` folders from VMware Fusion."""

    def __init__(self, path: Path, **kwargs):
        super().__init__(next(path.glob("*.vmx")), **kwargs)

    @staticmethod
    def detect(path: Path) -> bool:
        return path.is_dir() and path.suffix.lower() == ".vmwarevm" and len(list(path.glob("*.vmx"))) == 1



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.apps.webserver import apache, caddy, iis, nginx, webserver
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_access_logs_webserver_namespace_unix(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    apache_access_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    apache_error_file = absolute_path("_data/plugins/apps/webserver/apache/error.log")
    nginx_file = absolute_path("_data/plugins/apps/webserver/nginx/access.log")
    caddy_file = absolute_path("_data/plugins/apps/webserver/caddy/access.log")

    fs_unix.map_file("var/log/apache2/access.log", apache_access_file)
    fs_unix.map_file("var/log/apache2/error.log", apache_error_file)
    fs_unix.map_file("var/log/nginx/access.log", nginx_file)
    fs_unix.map_file("var/log/caddy_access.log", caddy_file)

    target_unix.add_plugin(apache.ApachePlugin)
    target_unix.add_plugin(nginx.NginxPlugin)
    target_unix.add_plugin(caddy.CaddyPlugin)

    # Register the IISLogsPlugin even though it is not compatible, to prevent it from being 'autodiscovered'
    target_unix.add_plugin(iis.IISLogsPlugin, check_compatible=False)
    target_unix.add_plugin(webserver.WebserverPlugin)

    access_logs = list(target_unix.webserver.access())

    assert len(access_logs) == 10
    assert len([record for record in access_logs if "apache" in record.source.as_posix()]) == 6
    assert len([record for record in access_logs if "nginx" in record.source.as_posix()]) == 2
    assert len([record for record in access_logs if "caddy" in record.source.as_posix()]) == 2

    error_logs = list(target_unix.webserver.error())
    assert len(error_logs) == 3

    # The logs function should yield both access and error logs
    assert len(list(target_unix.webserver.logs())) == 13


def test_access_logs_webserver_namespace_windows(target_win: Target, fs_win: VirtualFilesystem) -> None:
    config_path = absolute_path("_data/plugins/apps/webserver/iis/iis-applicationHost-iis.config")
    data_dir = absolute_path("_data/plugins/apps/webserver/iis/iis-logs-iis")

    fs_win.map_file("windows/system32/inetsrv/config/applicationHost.config", config_path)
    fs_win.map_dir("Users/John/iis-logs", data_dir)

    target_win.add_plugin(iis.IISLogsPlugin)
    target_win.add_plugin(webserver.WebserverPlugin)

    access_logs = list(target_win.webserver.access())
    assert len(access_logs) == 10



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.util import ts

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.plugin import Plugin, export
from dissect.target.plugins.os.unix.generic import calculate_last_activity

if TYPE_CHECKING:
    from datetime import datetime


class GenericPlugin(Plugin):
    """Generic FortiOS plugin."""

    def check_compatible(self) -> None:
        if self.target.os != "fortios":
            raise UnsupportedPluginError("FortiOS specific plugin loaded on non-FortiOS target")

    @export(property=True)
    def install_date(self) -> datetime | None:
        """Return the likely install date of FortiOS."""
        files = ["/data/etc/cloudinit.log", "/data/.vm_provisioned", "/data/etc/ssh/ssh_host_dsa_key"]
        for file in files:
            if (fp := self.target.fs.path(file)).exists():
                return ts.from_unix(fp.stat().st_mtime)
        return None

    @export(property=True)
    def activity(self) -> datetime | None:
        """Return last seen activity based on filesystem timestamps."""
        log_dirs = ["/var/log/log/root", "/var/log/root", "/data"]
        for log_dir in log_dirs:
            if (var_log := self.target.fs.path(log_dir)).exists():
                return calculate_last_activity(var_log)
        return None



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import (
    RegistryError,
    RegistryValueNotFoundError,
    UnsupportedPluginError,
)
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

ServiceRecord = TargetRecordDescriptor(
    "windows/service",
    [
        ("datetime", "ts"),
        ("string", "name"),
        ("string", "displayname"),
        ("string", "description"),
        ("path", "servicedll"),
        ("path", "imagepath"),
        ("string", "imagepath_args"),
        ("string", "objectname"),
        ("string", "start"),
        ("string", "type"),
        ("string", "errorcontrol"),
    ],
)

SERVICE_ENUMS = {
    "Type": {
        1: "Kernel Device Driver (0x1)",
        2: "File System Driver (0x2)",
        4: "Adapter (0x4)",
        16: "Service - Own Process (0x10)",
        32: "Service - Share Process (0x20)",
    },
    "Start": {
        0: "Boot (0)",
        1: "System (1)",
        2: "Auto Start (2)",
        3: "Manual (3)",
        4: "Disabled (4)",
    },
    "ErrorControl": {
        0: "Ignore (0)",
        1: "Normal (1)",
        2: "Severe (2)",
        3: "Critical (3)",
    },
}

# Make this general, use of other plugins
RE_PATH_SPLIT = re.compile(r"([\\/][a-zA-Z0-9_\.-]+\.[a-zA-Z0-9]{3})($| )")
RE_PATH_SPLIT_FALLBACK = re.compile(r"([\\/][a-zA-Z0-9_\.-]+)($| [^a-zA-Z0-9\\/])")


class ServicesPlugin(Plugin):
    """Services plugin."""

    KEY = "HKLM\\SYSTEM\\CurrentControlSet\\Services"

    def check_compatible(self) -> None:
        if not len(list(self.target.registry.keys(self.KEY))) > 0:
            raise UnsupportedPluginError("No services found in the registry")

    @export(record=ServiceRecord)
    def services(self) -> Iterator[ServiceRecord]:
        """Return information about all installed Windows services.

        The HKLM\\SYSTEM\\CurrentControlSet\\Services registry key contains information about the installed services and
        drivers on the system.

        References:
            - https://artifacts-kb.readthedocs.io/en/latest/sources/windows/ServicesAndDrivers.html

        Yields ServiceRecords with fields:

        .. code-block:: text

            hostname (string): The target hostname.
            domain (string): The target domain.
            ts (datatime): The last modified timestamp of the registry key.
            name (string): The service name.
            displayname (string): The service display name.
            description (string): The service description field.
            servicedll (path): The service dll.
            imagepath (path): The service image path.
            objectname (string): The object under which the service runs (for example LocalSystem)
            start (string): The service start field.
            type (string): The service type field.
            errorcontrol (string): The service error control field.
        """
        for controlset in self.target.registry.keys(self.KEY):
            for key in controlset.subkeys():
                ts = key.ts

                name = key.name
                servicedll = None

                try:
                    key.value("Type")
                except RegistryValueNotFoundError:
                    continue

                try:
                    servicedll = key.subkey("Parameters").value("ServiceDll").value
                    servicedll = self.target.fs.path(servicedll)
                except RegistryError:
                    pass

                image_path = None
                image_path_args = None

                display_name = None
                try:
                    display_name = key.value("DisplayName").value
                except RegistryValueNotFoundError:
                    pass

                description = None
                try:
                    description = key.value("Description").value
                except RegistryValueNotFoundError:
                    pass

                object_name = None
                try:
                    object_name = key.value("ObjectName").value
                except RegistryValueNotFoundError:
                    pass

                try:
                    image_path = key.value("ImagePath").value

                    if image_path:
                        if image_path.startswith('"'):
                            p = image_path.split('"', 2)

                            image_path = p[1]
                            image_path_args = p[2].strip()
                        else:
                            m = RE_PATH_SPLIT.search(image_path)
                            if not m:
                                m = RE_PATH_SPLIT_FALLBACK.search(image_path)

                            if m:
                                image_path_args = image_path[m.end(0) :]
                                image_path = image_path[: m.end(0)].strip()
                            else:
                                pass
                        image_path = self.target.fs.path(image_path)
                except RegistryError:
                    pass

                service_control = {}
                keys = ["Start", "Type", "ErrorControl"]
                for service_information in keys:
                    try:
                        value = key.value(service_information).value
                        value = SERVICE_ENUMS[service_information].get(value)
                        attr = f"service_{service_information.lower()}"
                        service_control[attr] = value
                    except RegistryError:  # noqa: PERF203
                        pass
                err_ctr = service_control.get("service_errorcontrol")
                yield ServiceRecord(
                    ts=ts,
                    name=name,
                    displayname=display_name,
                    description=description,
                    servicedll=servicedll,
                    imagepath=image_path,
                    imagepath_args=image_path_args,
                    objectname=object_name,
                    start=service_control.get("service_start"),
                    type=service_control.get("service_type"),
                    errorcontrol=err_ctr,
                    _target=self.target,
                )



from __future__ import annotations

import io
from typing import TYPE_CHECKING, BinaryIO

from dissect.util.stream import AlignedStream, BufferedStream

from dissect.target.container import Container

if TYPE_CHECKING:
    from pathlib import Path


class RawContainer(Container):
    __type__ = "raw"

    def __init__(self, fh: BinaryIO | Path, *args, **kwargs):
        if not hasattr(fh, "read"):
            fh = fh.open("rb")

        if not hasattr(fh, "size"):
            fh.seek(0, io.SEEK_END)
            size = fh.tell()
            fh.seek(0)
        elif callable(fh.size):
            size = fh.size()
        else:
            size = fh.size

        if not isinstance(fh, AlignedStream):
            fh = BufferedStream(fh, size=size)

        self.read = fh.read
        self.seek = fh.seek
        self.tell = fh.tell
        self.close = fh.close

        super().__init__(fh, size, *args, **kwargs)

    @staticmethod
    def _detect_fh(fh: BinaryIO, original: list | BinaryIO) -> bool:
        return True

    @staticmethod
    def detect_path(path: Path, original: list | BinaryIO) -> bool:
        return True



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

import pytest

from dissect.target.plugins.os.unix.bsd.darwin.macos.network import MacOSNetworkPlugin

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.fixture
def almost_empty_plist() -> dict:
    return {"CurrentSet": {}}


@pytest.fixture
def fake_plist() -> dict:
    return {
        "CurrentSet": "/Sets/1",
        "NetworkServices": {
            "1": {
                "DNS": {"ServerAddresses": ["8.8.8.8"]},
                "IPv4": {
                    "Addresses": ["192.122.13.34"],
                    "Router": "8.8.8.8",
                },
                "Interface": {
                    "DeviceName": "en0",
                    "Type": "Ethernet",
                },
            },
        },
        "Sets": {
            "1": {
                "Network": {
                    "Global": {"IPv4": {"ServiceOrder": ["1"]}},
                },
            },
        },
        "VirtualNetworkInterfaces": {"VLAN": {"vlan0": {"Tag": 2}}},
    }


@pytest.fixture
def vlan0(fake_plist: dict) -> dict:
    fake_plist["NetworkServices"]["1"]["Interface"].update({"DeviceName": "vlan0"})
    return fake_plist


@pytest.fixture
def inactive(fake_plist: dict) -> dict:
    fake_plist["NetworkServices"]["1"].update({"__INACTIVE__": True})
    return fake_plist


@pytest.fixture
def ipv6(fake_plist: dict) -> dict:
    del fake_plist["NetworkServices"]["1"]["IPv4"]
    fake_plist["NetworkServices"]["1"]["IPv6"] = {"Addresses": ["::1"]}
    return fake_plist


@pytest.fixture
def reorder(fake_plist: dict) -> dict:
    fake_plist["Sets"]["1"]["Network"]["Global"]["IPv4"]["ServiceOrder"] = ["2", "1"]
    return fake_plist


@pytest.fixture
def double(fake_plist: dict) -> dict:
    fake_plist["NetworkServices"]["2"] = fake_plist["NetworkServices"]["1"]
    return fake_plist


@pytest.fixture
def dhcp(fake_plist: dict) -> dict:
    fake_plist["NetworkServices"]["1"]["IPv4"].update({"ConfigMethod": "DHCP"})
    return fake_plist


@pytest.mark.parametrize(
    ("lease", "netinfo_param", "expected", "count"),
    [
        ({"IPAddress": None}, "almost_empty_plist", [], 0),
        (
            {},
            "fake_plist",
            [
                (0, "name", ["en0"]),
                (0, "type", ["Ethernet"]),
                (0, "enabled", ["True"]),
                (0, "cidr", ["192.122.13.34/32"]),
                (0, "hostname", ["dummys Mac"]),
                (0, "gateway", ["8.8.8.8"]),
                (0, "dns", ["8.8.8.8"]),
                (0, "mac", []),
                (0, "interface_service_order", ["0"]),
                (0, "vlan", ["None"]),
                (0, "domain", ["None"]),
            ],
            1,
        ),
        (
            {"IPAddress": "10.0.0.2"},
            "dhcp",
            [
                (0, "cidr", sorted(["10.0.0.2/32"])),
            ],
            1,
        ),
        (
            {},
            "vlan0",
            [
                (0, "vlan", ["2"]),
            ],
            1,
        ),
        (
            {},
            "inactive",
            [
                (0, "enabled", ["False"]),
            ],
            1,
        ),
        (
            {},
            "ipv6",
            [
                (0, "cidr", ["::1/128"]),
            ],
            1,
        ),
        (
            {},
            "reorder",
            [
                (0, "interface_service_order", ["1"]),
            ],
            1,
        ),
        (
            {},
            "double",
            [
                (0, "enabled", ["True"]),
                (1, "enabled", ["True"]),
            ],
            2,
        ),
    ],
)
def test_macos_network(
    target_macos: Target, lease: dict, netinfo_param: str, expected: dict, count: int, request: pytest.FixtureRequest
) -> None:
    plistnetwork = request.getfixturevalue(netinfo_param)
    with (
        mock.patch(
            "dissect.target.plugins.os.unix.bsd.darwin.macos.network.MacOSNetworkPlugin._plistlease", return_value=lease
        ),
        mock.patch(
            "dissect.target.plugins.os.unix.bsd.darwin.macos.network.MacOSNetworkPlugin._plistnetwork",
            return_value=plistnetwork,
        ),
    ):
        network = MacOSNetworkPlugin(target_macos)

    interfaces = list(network.interfaces())
    assert len(interfaces) == count
    for index, key, value in expected:
        attr = getattr(interfaces[index], key)
        if not isinstance(attr, list):
            attr = [attr]
        attr = sorted(map(str, attr))
        assert attr == value



from __future__ import annotations

import json
import logging
import re
from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util import ts

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.helpers.protobuf import ProtobufVarint
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import arg, export
from dissect.target.plugins.apps.container.container import (
    COMMON_CONTAINER_FIELDS,
    COMMON_IMAGE_FIELDS,
    COMMON_LOG_FIELDS,
    ContainerPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

log = logging.getLogger(__name__)

DockerContainerRecord = TargetRecordDescriptor(
    "apps/containers/docker/container",
    COMMON_CONTAINER_FIELDS,
)

DockerImageRecord = TargetRecordDescriptor(
    "apps/containers/docker/image",
    COMMON_IMAGE_FIELDS,
)

DockerLogRecord = TargetRecordDescriptor(
    "apps/containers/docker/log",
    COMMON_LOG_FIELDS,
)

# References:
# - https://github.com/moby/moby/pull/37092
# - https://github.com/cpuguy83/docker/blob/master/daemon/logger/local/doc.go
# - https://github.com/moby/moby/blob/master/api/types/plugins/logdriver/entry.proto
local_def = """
struct entry {
    uint32   header;

    // source
    uint8    s_type;        // 0x0a
    varint   s_len;         // 0x06
    char     source[s_len]; // stdout or stderr

    // timestamp
    uint8    t_type;        // 0x10
    varint   ts;            // timestamp in ums

    // message
    uint8    m_type;        // 0x1a
    varint   m_len;         // message length
    char     message[m_len];

    // partial_log_metadata not implemented

    uint32 footer;
};
"""

c_local = cstruct(endian=">")
c_local.add_custom_type("varint", ProtobufVarint, size=None, alignment=1, signed=False)
c_local.load(local_def, compiled=False)

RE_DOCKER_NS = re.compile(r"\.(?P<nanoseconds>\d{7,})(?P<postfix>Z|\+\d{2}:\d{2})")
RE_ANSI_ESCAPE = re.compile(r"\x1b(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

ASCII_MAP = {
    "\x08": "[BS]",
    "\x09": "[TAB]",
    "\x0a": "",  # \n
    "\x0d": "",  # \r
}


class DockerPlugin(ContainerPlugin):
    """Parse Docker Daemon artefacts.

    References:
        - https://didactic-security.com/resources/docker-forensics.pdf
        - https://didactic-security.com/resources/docker-forensics-cheatsheet.pdf
        - https://github.com/google/docker-explorer
    """

    __namespace__ = "docker"

    def __init__(self, target: Target):
        super().__init__(target)
        self.installs = set(find_installs(target))

    def check_compatible(self) -> None:
        if not self.installs:
            raise UnsupportedPluginError("No Docker install(s) found")

    @export(record=DockerImageRecord)
    def images(self) -> Iterator[DockerImageRecord]:
        """Returns any pulled docker images on the target system."""

        for data_root in self.installs:
            images_path = data_root.joinpath("image/overlay2/repositories.json")

            if not images_path.exists():
                self.target.log.debug("No docker images found, file %s does not exist", images_path)
                continue

            try:
                repositories = json.loads(images_path.read_text()).get("Repositories", {})
            except (json.JSONDecodeError, UnicodeDecodeError) as e:
                self.target.log.warning("Unable to parse JSON in: %s", images_path)
                self.target.log.debug("", exc_info=e)
                continue

            for name, tags in repositories.items():
                for tag, hash in tags.items():
                    image_metadata_path = data_root.joinpath(
                        "image/overlay2/imagedb/content/sha256/", hash.split(":")[-1]
                    )
                    created = None

                    if image_metadata_path.exists():
                        try:
                            image_metadata = json.loads(image_metadata_path.read_text())
                            created = convert_timestamp(image_metadata.get("created"))
                        except (json.JSONDecodeError, UnicodeDecodeError) as e:
                            self.target.log.warning("Unable to parse JSON in: %s", image_metadata_path)
                            self.target.log.debug("", exc_info=e)

                    yield DockerImageRecord(
                        name=name,
                        tag=tag,
                        image_id=hash_to_image_id(hash),
                        created=created,
                        hash=hash,
                        source=images_path,
                        _target=self.target,
                    )

    @export(record=DockerContainerRecord)
    def containers(self) -> Iterator[DockerContainerRecord]:
        """Returns any docker containers present on the target system."""

        for data_root in self.installs:
            for config_path in data_root.joinpath("containers").glob("**/config.v2.json"):
                try:
                    config = json.loads(config_path.read_text())
                except (json.JSONDecodeError, UnicodeDecodeError) as e:
                    self.target.log.warning("Unable to parse JSON in file: %s", config_path)
                    self.target.log.debug("", exc_info=e)
                    continue

                container_id = config.get("ID")

                # determine state
                running = config.get("State", {}).get("Running")
                if running:
                    ports = config.get("NetworkSettings", {}).get("Ports", {})

                if not running or not ports:
                    ports = config.get("Config", {}).get("ExposedPorts", {})

                # parse volumes
                volumes = []
                if mount_points := config.get("MountPoints", {}):
                    volumes = [
                        f"{mount_point.get('Source')}:{mount_point.get('Destination')}"
                        for mount_point in mount_points.values()
                    ]

                # determine mount point
                mount_path = None
                if container_id and config.get("Driver") == "overlay2":
                    mount_path = data_root.joinpath("image/overlay2/layerdb/mounts", container_id)
                    if not mount_path.exists():
                        self.target.log.warning("Overlay2 mount path does not exist for container: %s", container_id)

                else:
                    self.target.log.warning("Encountered unsupported container filesystem: %s", config.get("Driver"))

                yield DockerContainerRecord(
                    container_id=container_id,
                    image=config.get("Config", {}).get("Image"),
                    image_id=config.get("Image", "").split(":")[-1],
                    command=f"{config.get('Path', '')} {' '.join(config.get('Args', []))}".strip(),
                    created=convert_timestamp(config.get("Created")),
                    running=running,
                    pid=config.get("State", {}).get("Pid"),
                    started=convert_timestamp(config.get("State", {}).get("StartedAt")),
                    finished=convert_timestamp(config.get("State", {}).get("FinishedAt")),
                    ports=list(convert_ports(ports)),
                    names=config.get("Name", "").replace("/", "", 1),
                    volumes=volumes,
                    environment=config.get("Config", {}).get("Env", []),
                    mount_path=mount_path,
                    config_path=config_path,
                    source=config_path,
                    _target=self.target,
                )

    @export(record=DockerLogRecord)
    @arg(
        "--raw-messages",
        action="store_true",
        help="preserve ANSI escape sequences and trailing newlines from log messages",
    )
    @arg(
        "--remove-backspaces",
        action="store_true",
        help="alter messages by removing ASCII backspaces and the corresponding characters",
    )
    def logs(self, raw_messages: bool = False, remove_backspaces: bool = False) -> Iterator[DockerLogRecord]:
        """Returns log files (stdout/stderr) from Docker containers.

        The default Docker Daemon log driver is ``json-file``, which
        performs no log rotation. Another log driver is ``local`` and
        performs log rotation and compresses log files more efficiently.

        Eventually ``local`` will likely replace ``json-file`` as the
        default log driver.

        References:
            - https://docs.docker.com/config/containers/logging/configure/
            - https://docs.docker.com/config/containers/logging/json-file/
            - https://docs.docker.com/config/containers/logging/local/
        """

        for data_root in self.installs:
            containers_path = data_root.joinpath("containers")

            for log_file in containers_path.glob("**/*.log*"):
                container = log_file.parent

                # json log driver
                if "-json.log" in log_file.name:
                    for log_entry in self._parse_json_log(log_file):
                        yield DockerLogRecord(
                            ts=log_entry.get("time"),
                            container=container.name,  # container hash
                            stream=log_entry.get("stream"),
                            message=(
                                log_entry.get("log")
                                if raw_messages
                                else strip_log(log_entry.get("log"), remove_backspaces)
                            ),
                            source=log_file,
                            _target=self.target,
                        )

                # local log driver
                else:
                    for log_entry in self._parse_local_log(log_file):
                        yield DockerLogRecord(
                            ts=ts.from_unix_us(log_entry.ts // 1000),
                            container=container.parent.name,  # container hash
                            stream=log_entry.source,
                            message=(
                                log_entry.message if raw_messages else strip_log(log_entry.message, remove_backspaces)
                            ),
                            source=log_file,
                            _target=self.target,
                        )

    def _parse_local_log(self, path: Path) -> Iterator[c_local.entry]:
        fh = open_decompress(path, "rb")  # can be a .gz file

        while True:
            try:
                entry = c_local.entry(fh)
                if entry.header != entry.footer:
                    self.target.log.warning(
                        "Could not reliably parse log entry at offset %i in file %s."
                        "Entry could be parsed incorrectly. Please report this "
                        "issue as Docker's protobuf could have changed.",
                        fh.tell(),
                        path,
                    )
                yield entry
            except EOFError:  # noqa: PERF203
                break

    def _parse_json_log(self, path: Path) -> Iterator[dict]:
        for line in open_decompress(path, "rt"):
            try:
                entry = json.loads(line)
            except (json.JSONDecodeError, UnicodeDecodeError) as e:
                self.target.log.warning("Could not decode JSON line in file: %s", path)
                self.target.log.debug("", exc_info=e)
                continue
            yield entry


def get_data_path(path: Path) -> str | None:
    """Returns the configured Docker daemon data-root path."""
    try:
        config = json.loads(path.open("rt").read())
    except (json.JSONDecodeError, UnicodeDecodeError) as e:
        log.warning("Could not read JSON file: %s", path)
        log.debug(exc_info=e)
        return None

    return config.get("data-root")


def find_installs(target: Target) -> Iterator[Path]:
    """Attempt to find additional configured and existing Docker daemon data-root folders.

    References:
        - https://docs.docker.com/config/daemon/
    """

    default_data_paths = [
        # Linux
        "/var/lib/docker",
        "/var/snap/docker/common/var-lib-docker",
        # Windows
        "sysvol/ProgramData/docker",
    ]

    default_config_paths = [
        # Linux
        "/etc/docker/daemon.json",
        "/var/snap/docker/current/config/daemon.json",
        # Windows
        "sysvol/ProgramData/docker/config/daemon.json",
    ]

    user_config_paths = [
        # Docker Desktop (macOS/Windows/Linux)
        ".docker/daemon.json",
    ]

    for path in default_data_paths:
        if (path := target.fs.path(path)).exists():
            yield path

    for path in default_config_paths:
        if (config_file := target.fs.path(path)).exists():
            if not (data_path := get_data_path(config_file)):
                target.log.info("Unable to get data-root from docker daemon file %s", config_file)
                continue
            if (data_root_path := target.fs.path(data_path)).exists():
                yield data_root_path

    for path in user_config_paths:
        for user_details in target.user_details.all_with_home():
            if (config_file := user_details.home_path.joinpath(path)).exists():
                if not (data_path := get_data_path(config_file)):
                    target.log.info("Unable to get data-root from docker daemon file %s", config_file)
                    continue
                if (data_root_path := target.fs.path(data_path)).exists():
                    yield data_root_path


def convert_timestamp(timestamp: str | None) -> str | None:
    """Docker sometimes uses (unpadded) 9 digit nanosecond precision
    in their timestamp logs, eg. "2022-12-19T13:37:00.123456789Z".

    Python has no native %n nanosecond strptime directive, so we
    strip the last three digits from the timestamp to force
    compatbility with the 6 digit %f microsecond directive.
    """

    if not timestamp:
        return None

    timestamp_nanoseconds_plus_postfix = timestamp[19:]
    match = RE_DOCKER_NS.match(timestamp_nanoseconds_plus_postfix)

    # Timestamp does not have nanoseconds if there is no match.
    if not match:
        return timestamp

    # Take the first six digits and reconstruct the timestamp.
    match = match.groupdict()
    microseconds = match["nanoseconds"][:6]
    return f"{timestamp[:19]}.{microseconds}{match['postfix']}"


def convert_ports(ports: dict[str, list | dict]) -> Iterator[str]:
    """Depending on the state of the container (turned on or off) we
    can salvage forwarded ports for the container in different
    parts of the config.v2.json file.

    This function attempts to be agnostic and deals with
    "Ports" lists and "ExposedPorts" dicts.

    NOTE: This function makes a couple of assumptions. Feel free to
    improve this helper function.

    Returns an iterator of strings in the format ``0.0.0.0:1234->5678/tcp``.
    """

    for key, value in ports.items():
        if isinstance(value, list):
            for v in value:
                yield f"{v['HostIp']}:{v['HostPort']}->{key}"

        elif isinstance(value, dict):
            # NOTE: We make the assumption the default broadcast ip 0.0.0.0 was used.
            yield f"0.0.0.0:{key.split('/')[0]}->{key}"


def hash_to_image_id(hash: str) -> str:
    """Convert the hash to an abbrevated docker image id."""
    return hash.split(":")[-1][:12]


def strip_log(input: str | bytes, exc_backspace: bool = False) -> str:
    """Remove ANSI escape sequences from a given input string.

    Also translates ASCII codes such as backspaces to readable format.

    References:
        - https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797#general-ascii-codes
    """

    if isinstance(input, bytes):
        input = input.decode("utf-8", errors="backslashreplace")

    out = RE_ANSI_ESCAPE.sub("", input)

    if exc_backspace:
        out = _replace_backspace(out)

    for hex, name in ASCII_MAP.items():
        out = out.replace(hex, name)

    return out


def _replace_backspace(input: str) -> str:
    """Remove ANSI backspace characters (``\x08``) and 'replay' their effect on the rest of the string.

    For example, with the input ``123\x084``, the output would be ``124``.
    """
    out = ""
    for char in input:
        if char == "\x08":
            out = out[:-1]
        else:
            out += char
    return out



from __future__ import annotations

from io import BytesIO

import pytest

from dissect.target.plugins.os.unix.linux.fortios._keys import (
    KERNEL_KEY_MAP,
    AesKey,
    ChaCha20Key,
    ChaCha20Seed,
)
from dissect.target.plugins.os.unix.linux.fortios._os import (
    aes_decrypt,
    decrypt_rootfs,
    key_iv_for_kernel_hash,
)


def test_kernel_key_map() -> None:
    # Ensure that the kernel key map is valid
    for kernel_hash, key in KERNEL_KEY_MAP.items():
        # test if the kernel hash is a valid hex string
        assert bytes.fromhex(kernel_hash)
        # test if the kernel hash is a valid length (sha256)
        assert len(bytes.fromhex(kernel_hash)) == 32

        assert isinstance(key, (AesKey, ChaCha20Key, ChaCha20Seed))

        assert isinstance(key.key, bytes)
        assert len(key.key) == 32

        if hasattr(key, "iv"):
            assert isinstance(key.iv, bytes)
            assert len(key.iv) == 16


def test_key_iv_for_kernel_hash() -> None:
    # test FFW_1801F-v7.4.2.F-build2571-FORTINET (KDF)
    key = key_iv_for_kernel_hash("d719f7fd533d05efb872907cf3711d0d895750d288b856ce70fefecbd7ace482")
    assert key.key == bytes.fromhex("39ef9ceb4262b49252164a4558b14a9b006d91a5247f5c797af281fade2198a8")
    assert key.iv == bytes.fromhex("f30a9e100417e2c390b763d2be2f2d03")

    # test FFW_3980E-v7.0.14.M-build0601-FORTINET (static key + IV)
    key = key_iv_for_kernel_hash("a494ec1713ab75a5ab58a847f096951e2de7ba899bef1a9a88a9c94d8efc4749")
    assert key.key == bytes.fromhex("bb48ece8482e277f307479b8923796aed3b536e83f5fadc0d758b36192626762")
    assert key.iv == bytes.fromhex("5394687ae6c679a74c7901267dfb9bb3")

    # test unknown hash
    with pytest.raises(ValueError, match="No known decryption keys for kernel hash: .*"):
        key_iv_for_kernel_hash("12345")


def test_decrypt_rootfs() -> None:
    # test decrypt of rootfs.gz header (FWF_81F_2R_POE-v7.4.3.F-build2573-FORTINET)
    encrypted_rootfs_header = bytes.fromhex("3ccb 7d85 b9b0 4c8e 8c92 36d4 1d9c c48c")
    key = bytes.fromhex("b9c77cfca5c3f4fe543b5b861b5eeab61b0bfd23fa93f52f5cd428bb5567ec37")
    iv = bytes.fromhex("25c9578ca8d04f8c55009ae41657d7dd")
    fh = decrypt_rootfs(BytesIO(encrypted_rootfs_header), ChaCha20Key(key, iv))
    assert fh.read(16) == b"\x1f\x8b\x08\x00J\xd6\xbbe\x00\x03\xa4\xb6S\x900\x00"

    # test bad decrypt
    bad_key = ChaCha20Key(key[::-1], iv[::-1])
    with pytest.raises(ValueError, match="Failed to decrypt: No gzip magic header found."):
        decrypt_rootfs(BytesIO(encrypted_rootfs_header), bad_key)


def test_aes_decrypt() -> None:
    # encrypted FGT_VM64_AZURE-v7.6.0.F-build3401-FORTINET.out/rootfs.gz
    encrypted_rootfs_header = bytes.fromhex(
        """
    3cd7 cc80 3328 3d18 4601 d1de 6440 e41b
    bf60 0a4a eb2b e38f 4068 8123 37d9
    """
    )
    decrypted_rootfs_header = bytes.fromhex(
        """
    1f8b 0800 f33f a166 0003 a4b6 638c 2e50
    b3a8 f9b6 6ddb b66d dbd8 6ddb ddbb
        """
    )
    key = KERNEL_KEY_MAP.get("5a4c18b9118124049955caa29824188d92fc51741f9576f90ea7a9df082b7657")
    assert isinstance(key, AesKey)
    data = aes_decrypt(BytesIO(encrypted_rootfs_header), key)
    assert data == decrypted_rootfs_header



import plistlib
from collections.abc import Iterator
from pathlib import Path

from dissect.util.plist import NSKeyedArchiver

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import (
    COMMON_APPLICATION_FIELDS,
    TargetRecordDescriptor,
)
from dissect.target.plugin import OperatingSystem, Plugin, export

IOSApplicationRecord = TargetRecordDescriptor(
    "ios/application",
    COMMON_APPLICATION_FIELDS,
)


class IOSApplicationsPlugin(Plugin):
    """iOS applications plugin."""

    PATH = "/private/var/containers/Bundle/Application"

    def check_compatible(self) -> None:
        if self.target.os != OperatingSystem.IOS or not self.target.fs.path(self.PATH).exists():
            raise UnsupportedPluginError("Target is not iOS or no Bundle Application folder found")

    @export(record=IOSApplicationRecord)
    def applications(self) -> Iterator[IOSApplicationRecord]:
        """Yield installed iOS apps."""

        for app_dir in self.target.fs.path(self.PATH).glob("*"):
            if not app_dir.is_dir():
                continue

            metadata = {}
            if (metadata_file := app_dir.joinpath("iTunesMetadata.plist")).exists():
                metadata = plistlib.load(metadata_file.open("rb"))

            bundle = {}
            if (bundle_file := app_dir.joinpath("BundleMetadata.plist")).exists():
                bundle = NSKeyedArchiver(bundle_file.open("rb"))["root"]

            # System apps do not have a iTunesMetadata.plist file.
            info = {}
            if not metadata and (info_file := next(app_dir.glob("*.app/Info.plist"), None)):
                info = plistlib.load(info_file.open("rb"))
                if info.get("CFBundleIdentifier", "").startswith("com.apple."):
                    info["author"] = "Apple"
                    info["type"] = "system"

            # Ultimate fallback to application name
            if not info.get("CFBundleDisplayName") and not metadata.get("itemName"):
                metadata["itemName"] = next(app_dir.glob("*.app"), Path()).name.removesuffix(".app")

            yield IOSApplicationRecord(
                ts_installed=bundle.get("installDate"),
                name=metadata.get("itemName") or info.get("CFBundleDisplayName"),
                version=metadata.get("bundleShortVersionString") or info.get("CFBundleShortVersionString"),
                author=metadata.get("artistName") or info.get("author"),
                type=metadata.get("kind") or info.get("type"),
                path=app_dir,
                _target=self.target,
            )



from __future__ import annotations

import stat
from datetime import timedelta, timezone
from typing import TYPE_CHECKING, BinaryIO, Optional

from dissect.fat import exfat
from dissect.util.stream import RunlistStream
from dissect.util.ts import dostimestamp

from dissect.target.exceptions import FileNotFoundError, NotADirectoryError
from dissect.target.filesystem import Filesystem, FilesystemEntry
from dissect.target.helpers import fsutil

if TYPE_CHECKING:
    from collections.abc import Iterator

ExfatFileTree = tuple[exfat.c_exfat.FILE, dict[str, Optional["ExfatFileTree"]]]


class ExfatFilesystem(Filesystem):
    __type__ = "exfat"

    def __init__(self, fh: BinaryIO, *args, **kwargs):
        super().__init__(fh, *args, case_sensitive=False, alt_separator="\\", **kwargs)
        self.exfat = exfat.ExFAT(fh)
        self.cluster_size = self.exfat.cluster_size

    @staticmethod
    def _detect(fh: BinaryIO) -> bool:
        return fh.read(11)[3:] == b"EXFAT   "

    def get(self, path: str) -> ExfatFilesystemEntry:
        return ExfatFilesystemEntry(self, path, self._get_entry(path))

    def _get_entry(self, path: str, root: ExfatFileTree | None = None) -> ExfatFileTree:
        dirent = root if root is not None else self.exfat.files["/"]

        # Programmatically we will often use the `/` separator, so replace it
        # with the native path separator of exFAT `/` is an illegal character
        # in exFAT filenames, so it's safe to replace
        parts = path.replace("/", "\\").split("\\")

        for part in parts:
            if not part:
                continue

            file_tree = dirent[1]
            if file_tree is None:
                raise NotADirectoryError(f"Not a directory: {path}")

            for entry_name, entry_file_tree in file_tree.items():
                if entry_name.upper() == part.upper():
                    dirent = entry_file_tree
                    break
            else:
                raise FileNotFoundError(f"File not found: {path}")

        return dirent


class ExfatFilesystemEntry(FilesystemEntry):
    def __init__(
        self,
        fs: ExfatFilesystem,
        path: str,
        entry: ExfatFileTree,
    ):
        super().__init__(fs, path, entry)
        self.size = self.entry[0].stream.data_length
        self.cluster = self.entry[0].stream.location

    def get(self, path: str) -> ExfatFilesystemEntry:
        """Get a filesystem entry relative from the current one."""
        full_path = fsutil.join(self.path, path, alt_separator=self.fs.alt_separator)
        return ExfatFilesystemEntry(self.fs, full_path, self.fs._get_entry(path, self.entry))

    def open(self) -> BinaryIO:
        if self.entry[0].stream.flags.not_fragmented:
            runlist = self.fs.exfat.runlist(self.cluster, True, self.size)
        else:
            runlist = self.fs.exfat.runlist(self.cluster, False)
        return RunlistStream(self.fs.exfat.filesystem, runlist, self.size, self.fs.cluster_size)

    def _iterdir(self) -> Iterator[tuple[str, ExfatFileTree]]:
        if not self.is_dir():
            raise NotADirectoryError(self.path)

        for entry_name, entry_file_tree in self.entry[1].items():
            if entry_name in (".", ".."):
                continue
            yield (entry_name, entry_file_tree)

    def iterdir(self) -> Iterator[str]:
        """List the directory contents of a directory. Returns a generator of strings."""
        for entry_name, _ in self._iterdir():
            yield entry_name

    def scandir(self) -> Iterator[ExfatFilesystemEntry]:
        """List the directory contents of this directory. Returns a generator of filesystem entries."""
        for entry_name, entry_file_tree in self._iterdir():
            path = fsutil.join(self.path, entry_name, alt_separator=self.fs.alt_separator)
            yield ExfatFilesystemEntry(self.fs, path, entry_file_tree)

    def is_symlink(self) -> bool:
        """Return whether this entry is a link."""
        return False

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a directory."""
        return bool(self.entry[0].metadata.attributes.directory)

    def is_file(self, follow_symlinks: bool = True) -> bool:
        """Return whether this entry is a file."""
        return not self.is_dir()

    def stat(self, follow_symlinks: bool = True) -> fsutil.stat_result:
        return self.lstat()

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of this entry."""
        fe = self.entry[0]
        size = fe.stream.data_length
        addr = fe.stream.location

        # exfat stores additional offsets for ctime and mtime to get 10 millisecond precision
        fe = fe.metadata

        # all timestamps are recorded in local time. the utc offset (of the system generating the timestamp in question)
        # is recorded in the associated tz byte
        c_tz = self.fs.exfat._utc_timezone(fe.create_timezone)
        c_tzinfo = timezone(timedelta(minutes=c_tz["offset"]), c_tz["name"])
        m_tz = self.fs.exfat._utc_timezone(fe.modified_timezone)
        m_tzinfo = timezone(timedelta(minutes=m_tz["offset"]), m_tz["name"])
        a_tz = self.fs.exfat._utc_timezone(fe.access_timezone)
        a_tzinfo = timezone(timedelta(minutes=a_tz["offset"]), a_tz["name"])

        ctime = dostimestamp(fe.create_time, fe.create_offset).replace(tzinfo=c_tzinfo)
        mtime = dostimestamp(fe.modified_time, fe.modified_offset).replace(tzinfo=m_tzinfo)
        atime = dostimestamp(fe.access_time).replace(tzinfo=a_tzinfo)

        # mode, ino, dev, nlink, uid, gid, size, atime, mtime, ctime
        st_info = [
            (stat.S_IFDIR if self.is_dir() else stat.S_IFREG) | 0o777,
            addr,
            id(self.fs),
            0,
            0,
            0,
            size,
            atime.timestamp(),
            mtime.timestamp(),
            ctime.timestamp(),
        ]
        return fsutil.stat_result(st_info)



from __future__ import annotations

import io
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.vpn.openvpn import (
    OpenVPNClient,
    OpenVPNParser,
    OpenVPNPlugin,
    OpenVPNServer,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.filesystem import Filesystem
    from dissect.target.target import Target


def map_openvpn_configs(filesystem: Filesystem, target_dir: Path) -> None:
    client_config = absolute_path("_data/plugins/apps/vpn/openvpn/client.conf")
    server_config = absolute_path("_data/plugins/apps/vpn/openvpn/server.conf")
    filesystem.map_file(str(target_dir.joinpath("server.conf")), server_config)
    filesystem.map_file(str(target_dir.joinpath("server.ovpn")), server_config)
    filesystem.map_file(str(target_dir.joinpath("client.conf")), client_config)
    filesystem.map_file(str(target_dir.joinpath("client.ovpn")), client_config)


@pytest.mark.parametrize(
    ("target", "fs", "map_path"),
    [
        (
            "target_win_users",
            "fs_win",
            "Program Files/OpenVPN/config/helper",
        ),
        (
            "target_win_users",
            "fs_win",
            "Users/John/OpenVPN/config/helper",
        ),
        (
            "target_unix_users",
            "fs_unix",
            "etc/openvpn",
        ),
    ],
)
def test_openvpn_plugin(target: str, fs: str, map_path: str, request: pytest.FixtureRequest) -> None:
    target: Target = request.getfixturevalue(target)
    fs: Filesystem = request.getfixturevalue(fs)
    map_openvpn_configs(fs, fs.path(map_path))
    target.add_plugin(OpenVPNPlugin)
    records = list(target.openvpn.config())
    _verify_records(records)


def _verify_records(records: list[OpenVPNClient | OpenVPNServer]) -> None:
    assert len(records) == 4

    for record in records:
        if record.name == "server":
            assert record.name == "server"
            assert record.local.val.compressed == "0.0.0.0"
            assert record.port == 1194
            assert record.topology is None
            assert record.server == "10.8.0.0 255.255.255.0"
            assert record.ifconfig_pool_persist == "/var/log/openvpn/ipp.txt"
            assert record.pushed_options == [
                "route 192.168.10.0 255.255.255.0",
                "route 192.168.20.0 255.255.255.0",
                "route 192.168.30.0 255.255.255.0",
            ]
            assert record.client_to_client.value is False
            assert record.duplicate_cn.value is False
            assert record.proto == "udp"
            assert record.dev == "tun"
            assert "BEGIN CERTIFICATE" in record.ca
            assert "BEGIN CERTIFICATE" in record.cert
            assert record.key is None
            assert record.redacted_key
            assert record.tls_auth == "/etc/a ta.key"
            assert record.status == "/var/log/openvpn/openvpn-status.log"
            assert record.log is None
        else:
            # Client
            assert record.remote == [
                "my-server-1 1194",
                "my-server-2 1194",
                "my-server-3 1195",
                "my-server-4 11912",
            ]
            assert record.name == "client"
            assert record.proto == "udp"
            assert record.dev == "tun"
            assert record.ca == "ca.crt"
            assert record.cert == "client.crt"
            assert record.key == "client.key"
            assert not record.redacted_key
            assert record.tls_auth == "ta.key"
            assert record.status is None
            assert record.log is None


@pytest.mark.parametrize(
    ("data_string", "expected_data"),
    [
        (
            "<connection>\nroute data\n</connection>\n",
            {"connection": {"list_item0": {"route": "data"}}},
        ),
        (
            "<ca>\n----- BEGIN PRIVATE DATA -----\n</ca>",
            {
                "ca": "----- BEGIN PRIVATE DATA -----\n",
            },
        ),
    ],
)
def test_openvpn_config(data_string: str, expected_data: dict | list) -> None:
    parser = OpenVPNParser()
    parser.parse_file(io.StringIO(data_string))

    assert parser.parsed_data == expected_data



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.esedb import esedb, record, table
from dissect.esedb.exceptions import KeyNotFoundError
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import export
from dissect.target.plugins.apps.browser.browser import (
    GENERIC_DOWNLOAD_RECORD_FIELDS,
    GENERIC_HISTORY_RECORD_FIELDS,
    BrowserPlugin,
    try_idna,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.plugins.general.users import UserDetails
    from dissect.target.target import Target


class WebCache:
    """Class for opening and pre-processing IE WebCache file."""

    def __init__(self, target: Target, fh: BinaryIO):
        self.target = target
        self.db = esedb.EseDB(fh)

    def find_containers(self, name: str) -> Iterator[table.Table]:
        """Look up all ``ContainerId`` values for a given container name.

        Args:
            name: The container name to look up all container IDs of.

        Yields:
            All ``ContainerId`` values for the requested container name.
        """
        try:
            table = self.db.table("Containers")

            for container_record in table.records():
                if record_name := container_record.get("Name"):
                    record_name = record_name.rstrip("\00").lower()
                    if record_name == name.lower():
                        container_id = container_record.get("ContainerId")
                        yield self.db.table(f"Container_{container_id}")

        except KeyError as e:
            self.target.log.warning("Exception while parsing EseDB Containers table")
            self.target.log.debug("", exc_info=e)

    def _iter_records(self, name: str) -> Iterator[record.Record]:
        """Yield records from a Webcache container.

        Args:
            name: The container name.

        Yields:
            Records from specified Webcache container.
        """
        for container in self.find_containers(name):
            try:
                yield from container.records()
            except Exception as e:  # noqa: PERF203
                self.target.log.warning("Error iterating IExplore container %s", container)
                self.target.log.debug("", exc_info=e)
                continue

    def history(self) -> Iterator[record.Record]:
        """Yield records from the history webcache container."""
        yield from self._iter_records("history")

    def downloads(self) -> Iterator[record.Record]:
        """Yield records from the iedownload webcache container."""
        yield from self._iter_records("iedownload")

    def cookies(self) -> None:
        raise NotImplementedError("Cookies plugin is not implemented for Internet Explorer yet")


class InternetExplorerPlugin(BrowserPlugin):
    """Internet explorer browser plugin."""

    __namespace__ = "iexplore"

    DIRS = ("AppData/Local/Microsoft/Windows/WebCache",)

    CACHE_FILENAME = "WebCacheV01.dat"

    BrowserHistoryRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/ie/history", GENERIC_HISTORY_RECORD_FIELDS
    )

    BrowserDownloadRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
        "browser/ie/download", GENERIC_DOWNLOAD_RECORD_FIELDS
    )

    def __init__(self, target: Target):
        super().__init__(target)

        self.users_dirs: list[tuple[UserDetails, Path]] = []
        for user_details in self.target.user_details.all_with_home():
            for ie_dir in self.DIRS:
                cdir = user_details.home_path.joinpath(ie_dir)
                if not cdir.exists():
                    continue
                self.users_dirs.append((user_details.user, cdir))

    def check_compatible(self) -> None:
        if not len(self.users_dirs):
            raise UnsupportedPluginError("No Internet Explorer directories found")

    def _iter_cache(self) -> Iterator[WebCache]:
        """Yield open IE Webcache files.

        Args:
            filename: Name of the Webcache file.

        Yields:
            Open Webcache file.

        Raises:
            FileNoteFoundError: If the webcache file could not be found.
        """
        for user, cdir in self.users_dirs:
            cache_file = cdir.joinpath(self.CACHE_FILENAME)
            try:
                yield user, cache_file, WebCache(self.target, cache_file.open())
            except FileNotFoundError:
                self.target.log.warning("Could not find %s file: %s", self.CACHE_FILENAME, cache_file)

    @export(record=BrowserHistoryRecord)
    def history(self) -> Iterator[BrowserHistoryRecord]:
        """Return browser history records from Internet Explorer.

        Yields BrowserHistoryRecord with the following fields:

        .. code-block:: text

            ts (datetime): Visit timestamp.
            browser (string): The browser from which the records are generated from.
            id (string): Record ID.
            url (uri): History URL.
            title (string): Page title.
            description (string): Page description.
            host (string): Hostname.
            visit_type (varint): Visit type.
            visit_count (varint): Amount of visits.
            hidden (string): Hidden value.
            typed (boolean): Typed value.
            session (varint): Session value.
            from_visit (varint): Record ID of the "from" visit.
            from_url (uri): URL of the "from" visit.
            source: (path): The source file of the history record.
        """
        for user, cache_file, cache in self._iter_cache():
            for container_record in cache.history():
                if not container_record.get("Url"):
                    continue

                _, _, url = container_record.get("Url", "").rstrip("\x00").partition("@")

                ts = None
                if accessed_time := container_record.get("AccessedTime"):
                    ts = wintimestamp(accessed_time)

                yield self.BrowserHistoryRecord(
                    ts=ts,
                    browser="iexplore",
                    id=container_record.get("EntryId"),
                    url=try_idna(url),
                    title=None,
                    description=None,
                    host=None,
                    visit_type=None,
                    visit_count=container_record.get("AccessCount"),
                    hidden=None,
                    typed=None,
                    session=None,
                    from_visit=None,
                    from_url=None,
                    source=cache_file,
                    _target=self.target,
                    _user=user,
                )

    @export(record=BrowserDownloadRecord)
    def downloads(self) -> Iterator[BrowserDownloadRecord]:
        """Return browser downloads records from Internet Explorer.

        Yields BrowserDownloadRecord with the following fields:

        .. code-block:: text

            ts_start (datetime): Download start timestamp.
            ts_end (datetime): Download end timestamp.
            browser (string): The browser from which the records are generated from.
            id (string): Record ID.
            path (string): Download path.
            url (uri): Download URL.
            size (varint): Download file size.
            state (varint): Download state number.
            source: (path): The source file of the download record.
        """
        for user, cache_file, cache in self._iter_cache():
            for container_record in cache.downloads():
                down_path = None
                down_url = None
                ts_end = wintimestamp(container_record.AccessedTime) if container_record.AccessedTime else None

                try:
                    response_headers = container_record.ResponseHeaders.decode("utf-16-le", errors="ignore")
                    # Not used here, but [-6:-3] should give: ref_url, mime_type, temp_download_path
                    down_url, down_path = response_headers.split("\x00")[-3:-1]
                except (AttributeError, KeyNotFoundError) as e:
                    self.target.log.error("Error parsing response headers: %s", e)  # noqa: TRY400
                    self.target.log.debug("", exc_info=e)

                yield self.BrowserDownloadRecord(
                    ts_start=None,
                    ts_end=ts_end,
                    browser="iexplore",
                    id=container_record.EntryId,
                    path=self.target.fs.path(down_path) if down_path else None,
                    url=down_url,
                    size=None,
                    state=None,
                    source=cache_file,
                    _target=self.target,
                    _user=user,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.plugin import export
from dissect.target.plugins.os.windows.dpapi.keyprovider.defaultpassword.defaultpassword import (
    DefaultPasswordKeyProvider,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


class WinlogonDefaultPasswordKeyProviderPlugin(DefaultPasswordKeyProvider):
    """Windows Winlogon DefaultPassword key provider plugin."""

    __namespace__ = "dpapi.keyprovider.defaultpassword.winlogon"

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("Windows registry plugin not available on target")

    @export(output="yield")
    def keys(self) -> Iterator[tuple[str, str]]:
        """Yield Windows Winlogon DefaultPassword strings.

        Extracts plaintext ``DefaultPassword`` values from the ``Winlogon`` registry.

        References:
            - https://learn.microsoft.com/en-us/troubleshoot/windows-server/user-profiles-and-logon/turn-on-automatic-logon
        """

        for key in self.target.registry.keys("HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon"):
            try:
                yield self.__namespace__, key.value("DefaultPassword").value
            except RegistryValueNotFoundError:  # noqa: PERF203
                pass
